from  enum import Enum
class MSG_TYPE(Enum):
    # Messaging
    MSG_TYPE_CHUNK                      = 0 # A chunk of a message (used for classical chat)
    MSG_TYPE_CONTENT                   = 1 # A full message (for some personality the answer is sent in bulk)
    MSG_TYPE_CONTENT_INVISIBLE_TO_AI   = 2 # A full message (for some personality the answer is sent in bulk)
    MSG_TYPE_CONTENT_INVISIBLE_TO_USER = 3 # A full message (for some personality the answer is sent in bulk)

    # Thoughts
    MSG_TYPE_THOUGHT_CHUNK             = 4 # A chunk of a thought content (used for classical chat)
    MSG_TYPE_THOUGHT_CONTENT           = 5 # A full thought content (for some personality the answer is sent in bulk)

    # Conditionning
    # Informations
    MSG_TYPE_EXCEPTION              = 6 # An exception occured
    MSG_TYPE_WARNING                = 7 # A warning occured
    MSG_TYPE_INFO                   = 8 # An information to be shown to user

    # Steps
    MSG_TYPE_STEP                   = 9 # An instant step (a step that doesn't need time to be executed)
    MSG_TYPE_STEP_START             = 10 # A step has started (the text contains an explanation of the step done by he personality)
    MSG_TYPE_STEP_PROGRESS          = 11 # The progress value (the text contains a percentage and can be parsed by the reception)
    MSG_TYPE_STEP_END               = 12# A step has been done (the text contains an explanation of the step done by he personality)

    #Extra
    MSG_TYPE_JSON_INFOS             = 13# A JSON output that is useful for summarizing the process of generation used by personalities like chain of thoughts and tree of thooughts
    MSG_TYPE_REF                    = 14# References (in form of  [text](path))
    MSG_TYPE_CODE                   = 15# A javascript code to execute
    MSG_TYPE_UI                     = 16# A vue.js component to show (we need to build some and parse the text to show it)

    #Commands
    MSG_TYPE_NEW_MESSAGE            = 17# A new message
    MSG_TYPE_FINISHED_MESSAGE       = 18# End of current message

    #Tool calling
    MSG_TYPE_TOOL_CALL              = 19# a tool call
    MSG_TYPE_TOOL_OUTPUT            = 20# the output of the tool

    MSG_TYPE_REASONING              = 21# the ai shows its reasoning
    MSG_TYPE_SCRATCHPAD             = 22# the ai shows its scratchpad
    MSG_TYPE_OBSERVATION            = 23# the ai shows its reasoning

    MSG_TYPE_ERROR                  = 24#a severe error hapened
    MSG_TYPE_GENERATING_TITLE_START = 25#title generation started
    MSG_TYPE_GENERATING_TITLE_END   = 26#title generation done

    MSG_TYPE_SOURCES_LIST           = 27# List of sources provided


class SENDER_TYPES(Enum):
    SENDER_TYPES_USER               = 0 # Sent by user
    SENDER_TYPES_AI                 = 1 # Sent by ai
    SENDER_TYPES_SYSTEM             = 2 # Sent by athe system


class SUMMARY_MODE(Enum):
    SUMMARY_MODE_SEQUENCIAL        = 0
    SUMMARY_MODE_HIERARCHICAL      = 0

class ELF_COMPLETION_FORMAT(Enum):
    Instruct = 0
    Chat = 1
    @classmethod
    def from_string(cls, format_string: str) -> 'ELF_COMPLETION_FORMAT':
        format_mapping = {
            "Instruct": cls.Instruct,
            "Chat": cls.Chat,
        }

        try:
            return format_mapping[format_string.upper()]
        except KeyError:
            raise ValueError(f"Invalid format string: {format_string}. Must be one of {list(format_mapping.keys())}.")
    
    def __str__(self):
        return self.name    


import requests
import json
import base64
import os
import mimetypes
import math
from typing import Optional, Callable, List, Union, Dict

import httpx
import tiktoken
import pipmaster as pm

from lollms_client.lollms_llm_binding import LollmsLLMBinding
from lollms_client.lollms_types import MSG_TYPE, ELF_COMPLETION_FORMAT
from lollms_client.lollms_discussion import LollmsDiscussion
from lollms_client.lollms_utilities import encode_image
from ascii_colors import ASCIIColors, trace_exception

# Ensure required packages are installed
pm.ensure_packages(["httpx", "tiktoken"])

BindingName = "OpenLLMBinding"


def _read_file_as_base64(path):
    with open(path, "rb") as f:
        return base64.b64encode(f.read()).decode("utf-8")


def _extract_markdown_path(s):
    s = s.strip()
    if s.startswith("[") and s.endswith(")"):
        lb, rb = s.find("["), s.find("]")
        if lb != -1 and rb != -1 and rb > lb:
            return s[lb + 1 : rb].strip()
    return s


def _guess_mime_from_name(name, default="image/jpeg"):
    mime, _ = mimetypes.guess_type(name)
    return mime or default


def _to_data_url(b64_str, mime):
    return f"data:{mime};base64,{b64_str}"


def normalize_image_input(img, default_mime="image/jpeg"):
    """
    Returns an OpenAI API‑compatible content block for an image.
    Accepts various input formats and converts them to a data URL.
    """
    if isinstance(img, str):
        # Handle path‑like strings or raw base64
        s = _extract_markdown_path(img)
        if os.path.exists(s):
            b64 = _read_file_as_base64(s)
            mime = _guess_mime_from_name(s, default_mime)
            url = _to_data_url(b64, mime)
        else:  # Assume it's a raw base64 string
            url = _to_data_url(s, default_mime)
        return {"type": "image_url", "image_url": {"url": url}}

    raise ValueError("Unsupported image input type for OpenLLM")


class OpenLLMBinding(LollmsLLMBinding):
    """OpenLLM‑specific binding implementation"""

    def __init__(self, **kwargs):
        """
        Initialize the OpenLLM binding.

        Args:
            host_address (str): URL of the OpenLLM server (e.g. ``http://localhost:3000``).
            model_name (str): Name of the model to use.
            service_key (str): Authentication token for the service (optional).
            verify_ssl_certificate (bool): Whether to verify SSL certificates.
        """
        super().__init__(BindingName, **kwargs)
        self.host_address = kwargs.get("host_address")
        self.model_name = kwargs.get("model_name")
        self.service_key = kwargs.get("service_key", os.getenv("OPENLLM_API_KEY"))
        self.verify_ssl_certificate = kwargs.get("verify_ssl_certificate", True)

        if not self.host_address:
            raise ValueError("OpenLLM host address is required.")

        # Build headers – only include Authorization if a key is actually provided
        headers = {"Content-Type": "application/json"}
        if self.service_key:
            headers["Authorization"] = f"Bearer {self.service_key}"
        else:
            ASCIIColors.warning(
                "No service key provided for OpenLLM. Requests will be made without Authorization header."
            )

        # Append /v1 to the base URL for OpenAI compatibility
        base_url = f"{self.host_address.rstrip('/')}/v1"

        self.client = httpx.Client(
            base_url=base_url,
            headers=headers,
            verify=self.verify_ssl_certificate,
            timeout=None,
        )

    # --------------------------------------------------------------------- #
    # Helper methods
    # --------------------------------------------------------------------- #
    def _build_request_params(self, messages: list, **kwargs) -> dict:
        """Construct the JSON payload expected by the OpenLLM /chat/completions endpoint."""
        params = {
            "model": kwargs.get("model", self.model_name),
            "messages": messages,
            "stream": kwargs.get("stream", True),
        }

        # Map Lollms parameters to OpenAI‑compatible fields
        if "n_predict" in kwargs and kwargs["n_predict"] is not None:
            params["max_tokens"] = kwargs["n_predict"]
        if "temperature" in kwargs and kwargs["temperature"] is not None:
            params["temperature"] = kwargs["temperature"]
        if "top_p" in kwargs and kwargs["top_p"] is not None:
            params["top_p"] = kwargs["top_p"]
        if "top_k" in kwargs and kwargs["top_k"] is not None:
            params["top_k"] = kwargs["top_k"]
        if "repeat_penalty" in kwargs and kwargs["repeat_penalty"] is not None:
            params["frequency_penalty"] = kwargs["repeat_penalty"]
        if "seed" in kwargs and kwargs["seed"] is not None:
            params["seed"] = kwargs["seed"]

        return params

    def _process_request(
        self,
        params: dict,
        stream: Optional[bool],
        streaming_callback: Optional[Callable[[str, MSG_TYPE], None]],
    ) -> Union[str, dict]:
        """Execute the request – handling both streaming and non‑streaming modes."""
        output = ""
        try:
            if stream:
                with self.client.stream(
                    "POST", "/chat/completions", json=params
                ) as response:
                    if response.status_code != 200:
                        err = response.read().decode("utf-8")
                        raise Exception(
                            f"API Error: {response.status_code} - {err}"
                        )

                    for line in response.iter_lines():
                        if not line:
                            continue
                        if line.startswith("data:"):
                            data_str = line[len("data:") :].strip()
                            if data_str == "[DONE]":
                                break
                            try:
                                chunk = json.loads(data_str)
                                if chunk.get("choices"):
                                    delta = chunk["choices"][0].get("delta", {})
                                    word = delta.get("content", "")
                                    if word:
                                        if streaming_callback:
                                            if not streaming_callback(
                                                word, MSG_TYPE.MSG_TYPE_CHUNK
                                            ):
                                                break
                                        output += word
                            except json.JSONDecodeError:
                                continue
            else:
                response = self.client.post("/chat/completions", json=params)
                if response.status_code != 200:
                    raise Exception(
                        f"API Error: {response.status_code} - {response.text}"
                    )
                data = response.json()
                output = data["choices"][0]["message"]["content"]
                if streaming_callback:
                    streaming_callback(output, MSG_TYPE.MSG_TYPE_CHUNK)

        except Exception as e:
            trace_exception(e)
            err_msg = f"An error occurred with the OpenLLM API: {e}"
            if streaming_callback:
                streaming_callback(err_msg, MSG_TYPE.MSG_TYPE_EXCEPTION)
            return {"status": "error", "message": err_msg}

        return output

    # --------------------------------------------------------------------- #
    # Public API required by LollmsLLMBinding
    # --------------------------------------------------------------------- #
    def generate_text(
        self,
        prompt: str,
        images: Optional[List[str]] = None,
        system_prompt: str = "",
        n_predict: Optional[int] = None,
        stream: Optional[bool] = None,
        temperature: float = 0.7,
        top_k: int = 40,
        top_p: float = 0.9,
        repeat_penalty: float = 1.1,
        streaming_callback: Optional[Callable[[str, MSG_TYPE], None]] = None,
        **kwargs,
    ) -> Union[str, dict]:
        """Generate text (or multimodal output) via OpenLLM."""
        messages = []
        if system_prompt:
            messages.append({"role": "system", "content": system_prompt})

        user_content = [{"type": "text", "text": prompt}]
        if images:
            for img in images:
                user_content.append(normalize_image_input(img))

        messages.append({"role": "user", "content": user_content})

        params = self._build_request_params(
            messages=messages,
            n_predict=n_predict,
            stream=stream,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            repeat_penalty=repeat_penalty,
            **kwargs,
        )
        return self._process_request(params, stream, streaming_callback)

    def generate_from_messages(
        self,
        messages: List[Dict],
        n_predict: Optional[int] = None,
        stream: Optional[bool] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        repeat_penalty: Optional[float] = None,
        streaming_callback: Optional[Callable[[str, MSG_TYPE], None]] = None,
        **kwargs,
    ) -> Union[str, dict]:
        """Generate from a pre‑formatted list of OpenAI‑compatible messages."""
        params = self._build_request_params(
            messages=messages,
            n_predict=n_predict,
            stream=stream,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            repeat_penalty=repeat_penalty,
            **kwargs,
        )
        return self._process_request(params, stream, streaming_callback)

    def chat(
        self,
        discussion: LollmsDiscussion,
        branch_tip_id: Optional[str] = None,
        n_predict: Optional[int] = None,
        stream: Optional[bool] = None,
        temperature: float = 0.7,
        top_k: int = 40,
        top_p: float = 0.9,
        repeat_penalty: float = 1.1,
        repeat_last_n: int = 64,
        seed: Optional[int] = None,
        n_threads: Optional[int] = None,
        ctx_size: Optional[int] = None,
        streaming_callback: Optional[Callable[[str, MSG_TYPE], None]] = None,
        think: Optional[bool] = False,
        reasoning_effort: Optional[bool] = "low",
        reasoning_summary: Optional[bool] = "auto",
        **kwargs,
    ) -> Union[str, dict]:
        """
        Conduct a chat session using a :class:`LollmsDiscussion` object.
        The discussion is exported in an OpenAI‑compatible format and then
        passed to :meth:`_process_request`.
        """
        messages = discussion.export("openai_chat", branch_tip_id)

        params = self._build_request_params(
            messages=messages,
            n_predict=n_predict,
            stream=stream,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            repeat_penalty=repeat_penalty,
            **kwargs,
        )
        return self._process_request(params, stream, streaming_callback)

    def list_models(self) -> List[Dict]:
        """Return a list of models known to the OpenLLM server."""
        models_info = []
        try:
            response = self.client.get("/models")
            
            if response.status_code != 200:
                ASCIIColors.error(
                    f"OpenLLM /v1/models returned status {response.status_code}. "
                    f"Response body: {response.text}"
                )
                try:
                    response.raise_for_status()
                except Exception as e:
                    trace_exception(e)
                return models_info  # Empty list due to error

            models_data = response.json().get("data", [])
            for model in models_data:
                models_info.append(
                    {
                        "model_name": model.get("id", "N/A"),
                        "owned_by": model.get("owned_by", "N/A"),
                        "created": model.get("created", "N/A"),
                        "context_length": "unknown", # Not a standard field in OpenAI spec
                    }
                )
        except Exception as e:
            ASCIIColors.error(
                f"Failed to list models from OpenLLM: {e.__class__.__name__}: {e}"
            )
            trace_exception(e)
        return models_info


    def _get_encoding(self, model_name: str | None = None):
        """Fallback to tiktoken for generic tokenisation."""
        try:
            return tiktoken.encoding_for_model(model_name or self.model_name)
        except KeyError:
            return tiktoken.get_encoding("cl100k_base")

    def tokenize(self, text: str) -> list[int]:
        return self._get_encoding().encode(text)

    def detokenize(self, tokens: list[int]) -> str:
        return self._get_encoding().decode(tokens)

    def count_tokens(self, text: str) -> int:
        return len(self.tokenize(text))

    def embed(self, text: str | List[str], **kwargs) -> List:
        """
        Obtain embeddings via the OpenLLM ``/embeddings`` endpoint.
        If a single string is supplied, a single embedding vector is returned;
        otherwise a list of vectors is returned.
        """
        embedding_model = kwargs.get("model", self.model_name)
        single_input = isinstance(text, str)
        inputs = [text] if single_input else list(text)
        
        try:
            payload = {"model": embedding_model, "input": inputs}
            response = self.client.post("/embeddings", json=payload)
            response.raise_for_status()
            data = response.json()
            
            embeddings = [item["embedding"] for item in data.get("data", [])]

            if single_input and embeddings:
                return embeddings[0]
            else:
                return embeddings
                
        except Exception as e:
            ASCIIColors.error(
                f"Failed to generate embeddings using model '{embedding_model}': {e}"
            )
            trace_exception(e)
            return []

    def get_model_info(self) -> dict:
        """Return basic information about the current binding configuration."""
        return {
            "name": self.binding_name,
            "version": pm.get_installed_version("openllm")
            if pm.is_installed("openllm")
            else "unknown",
            "host_address": self.host_address,
            "model_name": self.model_name,
            "supports_structured_output": False,
            "supports_vision": True, # Assuming vision support based on original code
        }

    def load_model(self, model_name: str) -> bool:
        """Select a model for subsequent calls."""
        self.model_name = model_name
        ASCIIColors.info(f"OpenLLM model set to: {model_name}")
        return True

    def ps(self):
        """Placeholder – OpenLLM does not expose a process‑list endpoint."""
        return []


# Ensure the class is treated as concrete (no remaining abstract methods)
OpenLLMBinding.__abstractmethods__ = set()


# bindings/claude/__init__.py
import base64
import os
from io import BytesIO
from pathlib import Path
from typing import Optional, Callable, List, Union, Dict
import json
import requests

from lollms_client.lollms_discussion import LollmsDiscussion, LollmsMessage
from lollms_client.lollms_llm_binding import LollmsLLMBinding
from lollms_client.lollms_types import MSG_TYPE
from ascii_colors import ASCIIColors, trace_exception

import pipmaster as pm

# Ensure the required packages are installed
pm.ensure_packages(["anthropic", "pillow", "tiktoken", "requests"])

import anthropic
from PIL import Image, ImageDraw
import tiktoken

BindingName = "ClaudeBinding"

# API Endpoint for model listing
ANTHROPIC_API_BASE_URL = "https://api.anthropic.com/v1"

# A hardcoded list to be used as a fallback if the API call fails
_FALLBACK_MODELS = [
    {'model_name': 'claude-3-7-sonnet-20250219', 'display_name': 'Claude 3.7 Sonnet', 'description': 'Most intelligent model with extended thinking capabilities.', 'owned_by': 'Anthropic'},
    {'model_name': 'claude-3-5-sonnet-20240620', 'display_name': 'Claude 3.5 Sonnet', 'description': 'Our most intelligent model, a new industry standard.', 'owned_by': 'Anthropic'},
    {'model_name': 'claude-3-opus-20240229', 'display_name': 'Claude 3 Opus', 'description': 'Most powerful model for highly complex tasks.', 'owned_by': 'Anthropic'},
    {'model_name': 'claude-3-sonnet-20240229', 'display_name': 'Claude 3 Sonnet', 'description': 'Ideal balance of intelligence and speed for enterprise workloads.', 'owned_by': 'Anthropic'},
    {'model_name': 'claude-3-haiku-20240307', 'display_name': 'Claude 3 Haiku', 'description': 'Fastest and most compact model for near-instant responsiveness.', 'owned_by': 'Anthropic'},
    {'model_name': 'claude-2.1', 'display_name': 'Claude 2.1', 'description': 'Legacy model with a 200K token context window.', 'owned_by': 'Anthropic'},
    {'model_name': 'claude-2.0', 'display_name': 'Claude 2.0', 'description': 'Legacy model.', 'owned_by': 'Anthropic'},
    {'model_name': 'claude-instant-1.2', 'display_name': 'Claude Instant 1.2', 'description': 'Legacy fast and light-weight model.', 'owned_by': 'Anthropic'},
]


# Helper to check if a string is a valid path to an image
def is_image_path(path_str: str) -> bool:
    try:
        p = Path(path_str)
        return p.is_file() and p.suffix.lower() in ['.png', '.jpg', '.jpeg', '.gif', '.bmp', '.webp']
    except Exception:
        return False

# Helper to get image media type
def get_media_type(image_path: Union[str, Path]) -> str:
    path = Path(image_path)
    ext = path.suffix.lower()
    if ext == ".jpg" or ext == ".jpeg":
        return "image/jpeg"
    elif ext == ".png":
        return "image/png"
    elif ext == ".gif":
        return "image/gif"
    elif ext == ".webp":
        return "image/webp"
    else:
        # A default or raise an error
        return "image/jpeg"

class ClaudeBinding(LollmsLLMBinding):
    """Anthropic Claude-specific binding implementation."""

    def __init__(self,
                 **kwargs
                 ):
        """
        Initialize the Claude binding.

        Args:
            model_name (str): Name of the Claude model to use.
            service_key (str): Anthropic API key.
        """
        super().__init__(BindingName, **kwargs)
        self.model_name = kwargs.get("model_name")
        self.service_key = kwargs.get("service_key")
        self._cached_models: Optional[List[Dict[str, str]]] = None

        if not self.service_key:
            self.service_key = os.getenv("ANTHROPIC_API_KEY")
        
        if not self.service_key:
            raise ValueError("Anthropic API key is required. Please set it via the 'service_key' parameter or the ANTHROPIC_API_KEY environment variable.")

        try:
            self.client = anthropic.Anthropic(api_key=self.service_key)
        except Exception as e:
            ASCIIColors.error(f"Failed to configure Anthropic client: {e}")
            self.client = None
            raise ConnectionError(f"Could not configure Anthropic client: {e}") from e

    def _construct_parameters(self,
                              temperature: float,
                              top_p: float,
                              top_k: int,
                              n_predict: int) -> Dict[str, any]:
        """Builds a parameters dictionary for the Claude API."""
        params = {}
        if temperature is not None: params['temperature'] = float(temperature)
        if top_p is not None: params['top_p'] = top_p
        if top_k is not None: params['top_k'] = top_k
        if n_predict is not None: params['max_tokens'] = n_predict
        return params

    def generate_text(self,
                     prompt: str,
                     images: Optional[List[str]] = None,
                     system_prompt: str = "",
                     n_predict: Optional[int] = 2048,
                     stream: Optional[bool] = False,
                     temperature: float = 0.7,
                     top_k: int = 40,
                     top_p: float = 0.9,
                     repeat_penalty: float = 1.1, # Not supported
                     repeat_last_n: int = 64,   # Not supported
                     seed: Optional[int] = None,      # Not supported
                     n_threads: Optional[int] = None, # Not applicable
                     ctx_size: int | None = None,     # Determined by model
                     streaming_callback: Optional[Callable[[str, MSG_TYPE], None]] = None,
                     split:Optional[bool]=False, # Not used in this direct method
                     user_keyword:Optional[str]="!@>user:", # Not used
                     ai_keyword:Optional[str]="!@>assistant:", # Not used
                     think: Optional[bool] = False,
                     reasoning_effort: Optional[str] = "low", # low, medium, high
                     reasoning_summary: Optional[bool] = False, # auto
                     ) -> Union[str, dict]:
        """
        Generate text using the Claude model.
        """
        if not self.client:
            return {"status": False, "error": "Anthropic client not initialized."}

        # Handling Thinking / Reasoning
        thinking_config = None
        if think:
            # Map reasoning_effort to budget_tokens
            budget = 1024 # default/low
            if reasoning_effort == "medium":
                budget = 8192
            elif reasoning_effort == "high":
                budget = 16000
            
            # Constraint: max_tokens (n_predict) must be > budget_tokens
            # If default n_predict (2048) is too low for reasoning, boost it.
            required_min_tokens = budget + 2048 # Buffer for output
            if n_predict is None or n_predict < required_min_tokens:
                n_predict = required_min_tokens
                ASCIIColors.info(f"Adjusting n_predict to {n_predict} to accommodate thinking budget of {budget}")

            thinking_config = {"type": "enabled", "budget_tokens": budget}
            # Temperature must be removed or handled differently when thinking is enabled? 
            # Anthropic API usually allows temperature with thinking, but strict 1.0 might be enforced by API for some models. 
            # We'll leave it unless it errors. Note: Some documentation says temp should be 1.0 or not present for reasoning models, 
            # but Claude 3.7 supports it. We will let the API handle it.

        api_params = self._construct_parameters(temperature, top_p, top_k, n_predict)
        if thinking_config:
            api_params["thinking"] = thinking_config
            # Ensure max_tokens is set in params (it is set by _construct_parameters via n_predict)

        message_content = []
        if prompt and prompt.strip():
            message_content.append({"type": "text", "text": prompt})

        if images:
            for image_data in images:
                try:
                    if is_image_path(image_data):
                        with open(image_data, "rb") as image_file:
                            b64_data = base64.b64encode(image_file.read()).decode('utf-8')
                        media_type = get_media_type(image_data)
                    else:
                        b64_data = image_data
                        media_type = "image/jpeg"
                    
                    message_content.append({
                        "type": "image",
                        "source": { "type": "base64", "media_type": media_type, "data": b64_data }
                    })
                except Exception as e:
                    error_msg = f"Failed to process image: {e}"
                    ASCIIColors.error(error_msg)
                    return {"status": False, "error": error_msg}

        if not message_content:
            if stream and streaming_callback:
                streaming_callback("", MSG_TYPE.MSG_TYPE_FINISHED_MESSAGE)
            return ""

        messages = [{"role": "user", "content": message_content}]
        full_response_text = ""

        request_args = {
            "model": self.model_name,
            "messages": messages,
            **api_params
        }
        if system_prompt and system_prompt.strip():
            request_args["system"] = system_prompt

        try:
            if stream:
                # Use raw stream iteration to catch thinking events
                with self.client.messages.stream(**request_args) as stream_response:
                    in_thinking_block = False
                    for event in stream_response:
                        if event.type == "content_block_start" and event.content_block.type == "thinking":
                            full_response_text += "<think>\n"
                            if streaming_callback:
                                streaming_callback("<think>\n", MSG_TYPE.MSG_TYPE_CHUNK)
                            in_thinking_block = True
                        elif event.type == "content_block_delta" and event.delta.type == "thinking_delta":
                            chunk = event.delta.thinking
                            full_response_text += chunk
                            if streaming_callback:
                                streaming_callback(chunk, MSG_TYPE.MSG_TYPE_CHUNK)
                        elif event.type == "content_block_stop" and in_thinking_block:
                            full_response_text += "\n</think>\n"
                            if streaming_callback:
                                streaming_callback("\n</think>\n", MSG_TYPE.MSG_TYPE_CHUNK)
                            in_thinking_block = False
                        elif event.type == "content_block_delta" and event.delta.type == "text_delta":
                            chunk = event.delta.text
                            full_response_text += chunk
                            if streaming_callback:
                                if not streaming_callback(chunk, MSG_TYPE.MSG_TYPE_CHUNK):
                                    break
                return full_response_text
            else:
                response = self.client.messages.create(**request_args)
                if response.stop_reason == "error":
                     return {"status": False, "error": f"API returned an error: {response.stop_reason}"}
                
                # Reconstruct full text including thinking
                output_parts = []
                for block in response.content:
                    if block.type == "thinking":
                        output_parts.append(f"<think>\n{block.thinking}\n</think>\n")
                    elif block.type == "text":
                        output_parts.append(block.text)
                
                return "".join(output_parts)

        except Exception as ex:
            error_message = f"An unexpected error occurred with Claude API: {str(ex)}"
            trace_exception(ex)
            return {"status": False, "error": error_message}

    def chat(self,
             discussion: LollmsDiscussion,
             branch_tip_id: Optional[str] = None,
             n_predict: Optional[int] = 2048,
             stream: Optional[bool] = False,
             temperature: float = 0.7,
             top_k: int = 40,
             top_p: float = 0.9,
             repeat_penalty: float = 1.1, # Not supported
             repeat_last_n: int = 64, # Not supported
             seed: Optional[int] = None, # Not supported
             n_threads: Optional[int] = None, # Not supported
             ctx_size: Optional[int] = None, # Not supported
             streaming_callback: Optional[Callable[[str, MSG_TYPE], None]] = None,
             think: Optional[bool] = False,
             reasoning_effort: Optional[str] = "low", # low, medium, high
             reasoning_summary: Optional[bool] = False, # auto
             ) -> Union[str, dict]:
        """
        Conduct a chat session with the Claude model using a LollmsDiscussion object.
        """
        if not self.client:
             return {"status": "error", "message": "Anthropic client not initialized."}

        system_prompt = discussion.system_prompt
        messages = discussion.get_messages(branch_tip_id)
        
        history = []
        for msg in messages:
            role = 'user' if msg.sender_type == "user" else 'assistant'
            content_parts = []
            if msg.content and msg.content.strip():
                content_parts.append({"type": "text", "text": msg.content})
            
            if msg.images:
                for file_path in msg.images:
                    if is_image_path(file_path):
                        try:
                            with open(file_path, "rb") as image_file:
                                b64_data = base64.b64encode(image_file.read()).decode('utf-8')
                            media_type = get_media_type(file_path)
                            content_parts.append({
                                "type": "image",
                                "source": { "type": "base64", "media_type": media_type, "data": b64_data }
                            })
                        except Exception as e:
                            ASCIIColors.warning(f"Could not load image {file_path}: {e}")
            
            if content_parts:
                if history and history[-1]['role'] == role:
                    history[-1]['content'].extend(content_parts)
                else:
                    history.append({'role': role, 'content': content_parts})
        
        if not history:
            return {"status": "error", "message": "Cannot start chat with an empty discussion."}

        # Handling Thinking / Reasoning
        thinking_config = None
        if think:
            budget = 1024
            if reasoning_effort == "medium":
                budget = 8192
            elif reasoning_effort == "high":
                budget = 16000
            
            required_min_tokens = budget + 2048
            if n_predict is None or n_predict < required_min_tokens:
                n_predict = required_min_tokens
                ASCIIColors.info(f"Adjusting n_predict to {n_predict} for thinking budget {budget}")

            thinking_config = {"type": "enabled", "budget_tokens": budget}

        api_params = self._construct_parameters(temperature, top_p, top_k, n_predict)
        if thinking_config:
            api_params["thinking"] = thinking_config

        full_response_text = ""

        request_args = {
            "model": self.model_name,
            "messages": history,
            **api_params
        }
        if system_prompt and system_prompt.strip():
            request_args["system"] = system_prompt

        try:
            if stream:
                with self.client.messages.stream(**request_args) as stream_response:
                    in_thinking_block = False
                    for event in stream_response:
                        if event.type == "content_block_start" and event.content_block.type == "thinking":
                            full_response_text += "<think>\n"
                            if streaming_callback: streaming_callback("<think>\n", MSG_TYPE.MSG_TYPE_CHUNK)
                            in_thinking_block = True
                        elif event.type == "content_block_delta" and event.delta.type == "thinking_delta":
                            chunk = event.delta.thinking
                            full_response_text += chunk
                            if streaming_callback: streaming_callback(chunk, MSG_TYPE.MSG_TYPE_CHUNK)
                        elif event.type == "content_block_stop" and in_thinking_block:
                            full_response_text += "\n</think>\n"
                            if streaming_callback: streaming_callback("\n</think>\n", MSG_TYPE.MSG_TYPE_CHUNK)
                            in_thinking_block = False
                        elif event.type == "content_block_delta" and event.delta.type == "text_delta":
                            chunk = event.delta.text
                            full_response_text += chunk
                            if streaming_callback:
                                if not streaming_callback(chunk, MSG_TYPE.MSG_TYPE_CHUNK):
                                    break
                return full_response_text
            else:
                response = self.client.messages.create(**request_args)
                if response.stop_reason == "error":
                     return {"status": "error", "message": f"API returned an error: {response.stop_reason}"}
                
                output_parts = []
                for block in response.content:
                    if block.type == "thinking":
                        output_parts.append(f"<think>\n{block.thinking}\n</think>\n")
                    elif block.type == "text":
                        output_parts.append(block.text)
                return "".join(output_parts)

        except Exception as ex:
            error_message = f"An unexpected error occurred with Claude API: {str(ex)}"
            trace_exception(ex)
            return {"status": "error", "message": error_message}
            
    def tokenize(self, text: str) -> list:
        """
        Tokenize the input text.
        Note: Claude doesn't expose a public tokenizer API.
        Using tiktoken for a rough estimate, NOT accurate for Claude.
        """
        try:
            encoding = tiktoken.get_encoding("cl100k_base")
            return encoding.encode(text)
        except:
            return list(text.encode('utf-8'))

    def detokenize(self, tokens: list) -> str:
        """
        Detokenize a list of tokens.
        Note: Based on the placeholder tokenizer.
        """
        try:
            encoding = tiktoken.get_encoding("cl100k_base")
            return encoding.decode(tokens)
        except:
            return bytes(tokens).decode('utf-8', errors='ignore')

    def count_tokens(self, text: str) -> int:
        """
        Count tokens from a text using the Anthropic API.
        This provides a more accurate count than a fallback tokenizer.
        """
        if not text or not text.strip():
            return 0

        if not self.client:
            ASCIIColors.warning("Cannot count tokens, Anthropic client not initialized.")
            return len(self.tokenize(text))
        try:
            # Note: count_tokens doesn't use a system prompt, so it's safe.
            # However, for consistency, we could add one if needed by the logic.
            # For now, this is fine as it only counts user content tokens.
            response = self.client.messages.count_tokens( # Changed from messages.count_tokens to top-level client method
                model=self.model_name,
                messages=[{"role": "user", "content": text}]
            )
            return response.input_tokens # Updated to correct response attribute (it's usually 'input_tokens' in CountTokensResponse)
        except Exception as e:
            trace_exception(e)
            ASCIIColors.error(f"Failed to count tokens with Claude API: {e}")
            return len(self.tokenize(text))

    def embed(self, text: str, **kwargs) -> List[float]:
        """
        Get embeddings for the input text.
        Note: Anthropic does not provide a dedicated embedding model API.
        """
        ASCIIColors.warning("Anthropic does not offer a public embedding API. This method is not implemented.")
        raise NotImplementedError("Claude binding does not support embeddings.")

    def get_model_info(self) -> dict:
        """Return information about the current Claude model setup."""
        return {
            "name": self.binding_name,
            "version": anthropic.__version__,
            "host_address": "https://api.anthropic.com",
            "model_name": self.model_name,
            "supports_structured_output": False,
            "supports_vision": "claude-3" in self.model_name,
        }

    def list_models(self) -> List[Dict[str, str]]:
        """
        Lists available models from the Anthropic API.
        Caches the result to avoid repeated API calls.
        Falls back to a static list if the API call fails.
        """
        if self._cached_models is not None:
            return self._cached_models

        if not self.service_key:
            ASCIIColors.warning("Cannot fetch models without an API key. Using fallback list.")
            self._cached_models = _FALLBACK_MODELS
            return self._cached_models

        headers = {
            "x-api-key": self.service_key,
            "anthropic-version": "2023-06-01", 
            "accept": "application/json"
        }
        url = f"{ANTHROPIC_API_BASE_URL}/models"
        
        try:
            response = requests.get(url, headers=headers, timeout=15)
            response.raise_for_status()
            
            data = response.json()
            
            if "data" in data and isinstance(data["data"], list):
                models_data = data["data"]
                formatted_models = []
                for model in models_data:
                    model_id = model.get("id")
                    if not model_id: continue
                    
                    display_name = model.get("name", model_id.replace("-", " ").title())
                    
                    desc_parts = []
                    if model.get('context_length'):
                        desc_parts.append(f"Context: {model['context_length']:,} tokens.")
                    if model.get('max_output_tokens'):
                         desc_parts.append(f"Max Output: {model['max_output_tokens']:,} tokens.")
                    description = " ".join(desc_parts) or f"Anthropic model: {model_id}"

                    formatted_models.append({
                        'model_name': model_id,
                        'display_name': display_name,
                        'description': description,
                        'owned_by': 'Anthropic'
                    })

                formatted_models.sort(key=lambda x: x['model_name'])
                self._cached_models = formatted_models
                ASCIIColors.green(f"Successfully fetched and parsed {len(self._cached_models)} models.")
                return self._cached_models
            else:
                raise ValueError("API response is malformed. 'data' field missing or not a list.")

        except Exception as e:
            ASCIIColors.error(f"Failed to fetch models from Anthropic API: {e}")
            ASCIIColors.warning("Using hardcoded fallback list of models.")
            trace_exception(e)
            self._cached_models = _FALLBACK_MODELS
            return self._cached_models

    def load_model(self, model_name: str) -> bool:
        """Set the model name for subsequent operations."""
        self.model_name = model_name
        ASCIIColors.info(f"Claude model set to: {model_name}. It will be used on the next API call.")
        return True

if __name__ == '__main__':
    # Example Usage (requires ANTHROPIC_API_KEY environment variable)
    if 'ANTHROPIC_API_KEY' not in os.environ:
        ASCIIColors.red("Error: ANTHROPIC_API_KEY environment variable not set.")
        print("Please get your key from Anthropic and set it.")
        exit(1)

    ASCIIColors.yellow("--- Testing ClaudeBinding ---")

    # --- Configuration ---
    test_model_name = "claude-3-7-sonnet-20250219" # Use Haiku for speed in testing
    test_vision_model_name = "claude-3-5-sonnet-20240620"
    
    full_streamed_text = ""

    try:
        # --- Initialization ---
        ASCIIColors.cyan("\n--- Initializing Binding ---")
        binding = ClaudeBinding(model_name=test_model_name)
        ASCIIColors.green("Binding initialized successfully.")
        ASCIIColors.info(f"Using anthropic version: {anthropic.__version__}")

        # --- List Models ---
        ASCIIColors.cyan("\n--- Listing Models (dynamic) ---")
        models = binding.list_models()
        if models:
            ASCIIColors.green(f"Found {len(models)} models.")
            for m in models:
                print(f"- {m['model_name']} ({m['display_name']})")
        else:
            ASCIIColors.error("Failed to list models.")

        # --- Count Tokens ---
        ASCIIColors.cyan("\n--- Counting Tokens (with valid and empty text) ---")
        sample_text = "Hello, world! This is a test."
        token_count = binding.count_tokens(sample_text)
        ASCIIColors.green(f"Token count for '{sample_text}': {token_count} (via API)")
        empty_token_count = binding.count_tokens("  ")
        ASCIIColors.green(f"Token count for empty string: {empty_token_count} (handled locally)")
        assert empty_token_count == 0

        # --- Text Generation (Non-Streaming) ---
        ASCIIColors.cyan("\n--- Text Generation (Non-Streaming) ---")
        prompt_text = "Explain the importance of bees in one paragraph."
        ASCIIColors.info(f"Prompt: {prompt_text}")
        generated_text = binding.generate_text(prompt_text, n_predict=100, stream=False, system_prompt=" ", think=True)
        if isinstance(generated_text, str):
            ASCIIColors.green(f"Generated text:\n{generated_text}")
        else:
            ASCIIColors.error(f"Generation failed: {generated_text}")

        # --- Text Generation (Streaming) ---
        ASCIIColors.cyan("\n--- Text Generation (Streaming) ---")
        
        captured_chunks = []
        def stream_callback(chunk: str, msg_type: int):
            ASCIIColors.green(chunk, end="", flush=True)
            captured_chunks.append(chunk)
            return True
        
        ASCIIColors.info(f"Prompt: {prompt_text}")
        result = binding.generate_text(prompt_text, n_predict=150, stream=True, streaming_callback=stream_callback, think=True)
        full_streamed_text = "".join(captured_chunks)
        print("\n--- End of Stream ---")
        ASCIIColors.green(f"Full streamed text (for verification): {result}")
        assert result == full_streamed_text

        # --- Embeddings ---
        ASCIIColors.cyan("\n--- Embeddings ---")
        try:
            embedding_text = "Lollms is a cool project."
            embedding_vector = binding.embed(embedding_text)
        except NotImplementedError as e:
            ASCIIColors.green(f"Successfully caught expected error for embeddings: {e}")
        except Exception as e:
            ASCIIColors.error(f"Caught an unexpected error for embeddings: {e}")

        # --- Vision Model Test ---
        dummy_image_path = "claude_dummy_test_image.png"
        try:
            available_model_names = [m['model_name'] for m in models]
            if test_vision_model_name not in available_model_names:
                 ASCIIColors.warning(f"Vision test model '{test_vision_model_name}' not available. Skipping vision test.")
            else:
                img = Image.new('RGB', (200, 50), color = ('blue'))
                d = ImageDraw.Draw(img)
                d.text((10,10), "Test Image", fill=('yellow'))
                img.save(dummy_image_path)
                ASCIIColors.info(f"Created dummy image: {dummy_image_path}")

                ASCIIColors.cyan(f"\n--- Vision Generation (using {test_vision_model_name}) ---")
                binding.load_model(test_vision_model_name)
                vision_prompt = "What color is the text and what does it say?"
                ASCIIColors.info(f"Vision Prompt: {vision_prompt} with image {dummy_image_path}")
                
                vision_response = binding.generate_text(
                    prompt=vision_prompt,
                    images=[dummy_image_path],
                    n_predict=50,
                    stream=False
                )
                if isinstance(vision_response, str):
                    ASCIIColors.green(f"Vision model response: {vision_response}")
                else:
                    ASCIIColors.error(f"Vision generation failed: {vision_response}")
        except Exception as e:
            ASCIIColors.error(f"Error during vision test: {e}")
            trace_exception(e)
        finally:
            if os.path.exists(dummy_image_path):
                os.remove(dummy_image_path)

    except Exception as e:
        ASCIIColors.error(f"An error occurred during testing: {e}")
        trace_exception(e)

    ASCIIColors.yellow("\nClaudeBinding test finished.")


# bindings/lollms/binding.py
import requests
from lollms_client.lollms_llm_binding import LollmsLLMBinding
from lollms_client.lollms_types import MSG_TYPE
from lollms_client.lollms_utilities import encode_image
from lollms_client.lollms_types import ELF_COMPLETION_FORMAT
from lollms_client.lollms_discussion import LollmsDiscussion
from ascii_colors import ASCIIColors, trace_exception
from typing import Optional, Callable, List, Union
import json

BindingName = "LollmsWebuiLLMBinding"


class LollmsWebuiLLMBinding(LollmsLLMBinding):
    """LOLLMS-specific binding implementation"""
    
    DEFAULT_HOST_ADDRESS = "http://localhost:9600"
    
    def __init__(self, 
                 personality: Optional[int] = None, 
                 **kwargs
                 ):
        """
        Initialize the LOLLMS binding.

        Args:
            host_address (str): Host address for the LOLLMS service. Defaults to DEFAULT_HOST_ADDRESS.
            model_name (str): Name of the model to use. Defaults to empty string.
            service_key (str): Authentication key for the service. Defaults to None.
            verify_ssl_certificate (bool): Whether to verify SSL certificates. Defaults to True.
            personality (Optional[int]): Personality ID for generation. Defaults to None.
        """
        super().__init__(BindingName, **kwargs)
        host_address = kwargs.get("host_address")
        self.host_address=host_address if host_address is not None else self.DEFAULT_HOST_ADDRESS
        self.model_name=kwargs.get("model_name")
        self.service_key=kwargs.get("service_key")
        self.verify_ssl_certificate=kwargs.get("verify_ssl_certificate")
        self.default_completion_format=kwargs.get("default_completion_format",ELF_COMPLETION_FORMAT.Chat) 
        self.personality = kwargs.get("personality")
        self.model = None
    
    def generate_text(self,
                     prompt: str,
                     images: Optional[List[str]] = None,
                     system_prompt: str = "",
                     n_predict: Optional[int] = None,
                     stream: Optional[bool] = None,
                     temperature: Optional[float] = None,
                     top_k: Optional[int] = None,
                     top_p: Optional[float] = None,
                     repeat_penalty: Optional[float] = None,
                     repeat_last_n: Optional[int] = None,
                     seed: Optional[int] = None,
                     n_threads: Optional[int] = None,
                     ctx_size: int | None = None,
                     streaming_callback: Optional[Callable[[str, MSG_TYPE], None]] = None,
                     split:Optional[bool]=False, # put to true if the prompt is a discussion
                     user_keyword:Optional[str]="!@>user:",
                     ai_keyword:Optional[str]="!@>assistant:",
                     ) -> Union[str, dict]:
        """
        Generate text using the active LLM binding, using instance defaults if parameters are not provided.

        Args:
            prompt (str): The input prompt for text generation.
            images (Optional[List[str]]): List of image file paths for multimodal generation.
            n_predict (Optional[int]): Maximum number of tokens to generate. Uses instance default if None.
            stream (Optional[bool]): Whether to stream the output. Uses instance default if None.
            temperature (Optional[float]): Sampling temperature. Uses instance default if None.
            top_k (Optional[int]): Top-k sampling parameter. Uses instance default if None.
            top_p (Optional[float]): Top-p sampling parameter. Uses instance default if None.
            repeat_penalty (Optional[float]): Penalty for repeated tokens. Uses instance default if None.
            repeat_last_n (Optional[int]): Number of previous tokens to consider for repeat penalty. Uses instance default if None.
            seed (Optional[int]): Random seed for generation. Uses instance default if None.
            n_threads (Optional[int]): Number of threads to use. Uses instance default if None.
            ctx_size (int | None): Context size override for this generation.
            streaming_callback (Optional[Callable[[str, str], None]]): Callback function for streaming output.
                - First parameter (str): The chunk of text received.
                - Second parameter (str): The message type (e.g., MSG_TYPE.MSG_TYPE_CHUNK).
            split:Optional[bool]: put to true if the prompt is a discussion
            user_keyword:Optional[str]: when splitting we use this to extract user prompt 
            ai_keyword:Optional[str]": when splitting we use this to extract ai prompt

        Returns:
            Union[str, dict]: Generated text or error dictionary if failed.
        """
        # Determine endpoint based on presence of images
        endpoint = "/lollms_generate_with_images" if images else "/lollms_generate"
        url = f"{self.host_address}{endpoint}"
        
        # Set headers
        headers = {
            'Content-Type': 'application/json',
        }
        if self.service_key:
            headers['Authorization'] = f'Bearer {self.service_key}'

        # Handle images if provided
        image_data = []
        if images:
            for image_path in images:
                try:
                    encoded_image = encode_image(image_path)
                    image_data.append(encoded_image)
                except Exception as e:
                    return {"status": False, "error": f"Failed to process image {image_path}: {str(e)}"}
        
        # Prepare request data
        data = {
            "prompt":"!@>system: "+system_prompt+"\n"+"!@>user: "+prompt if system_prompt else prompt,
            "model_name": self.model_name,
            "personality": self.personality,
            "n_predict": n_predict,
            "stream": stream,
            "temperature": temperature,
            "top_k": top_k,
            "top_p": top_p,
            "repeat_penalty": repeat_penalty,
            "repeat_last_n": repeat_last_n,
            "seed": seed,
            "n_threads": n_threads
        }
        
        if image_data:
            data["images"] = image_data

        # Make the request
        response = requests.post(
            url, 
            json=data, 
            headers=headers, 
            stream=stream,
            verify=self.verify_ssl_certificate
        )
        
        if not stream:
            if response.status_code == 200:
                try:
                    text = response.text.strip()
                    return text
                except Exception as ex:
                    return {"status": False, "error": str(ex)}
            else:
                return {"status": False, "error": response.text}
        else:
            text = ""
            if response.status_code == 200:
                try:
                    for line in response.iter_lines():
                        chunk = line.decode("utf-8")
                        text += chunk
                        if streaming_callback:
                            streaming_callback(chunk, MSG_TYPE.MSG_TYPE_CHUNK)
                    # Handle potential quotes from streaming response
                    if text and text[0] == '"':
                        text = text[1:]
                    if text and text[-1] == '"':
                        text = text[:-1]
                    return text.rstrip('!')
                except Exception as ex:
                    return {"status": False, "error": str(ex)}
            else:
                return {"status": False, "error": response.text}
    def chat(self,
             discussion: LollmsDiscussion,
             branch_tip_id: Optional[str] = None,
             n_predict: Optional[int] = None,
             stream: Optional[bool] = None,
             temperature: Optional[float] = None,
             top_k: Optional[int] = None,
             top_p: Optional[float] = None,
             repeat_penalty: Optional[float] = None,
             repeat_last_n: Optional[int] = None,
             seed: Optional[int] = None,
             n_threads: Optional[int] = None,
             ctx_size: int | None = None,
             streaming_callback: Optional[Callable[[str, MSG_TYPE], None]] = None
             ) -> Union[str, dict]:
        """
        Conduct a chat session with a lollms-webui server using a LollmsDiscussion object.

        Args:
            discussion (LollmsDiscussion): The discussion object containing the conversation history.
            branch_tip_id (Optional[str]): The ID of the message to use as the tip of the conversation branch. Defaults to the active branch.
            ... (other parameters) ...

        Returns:
            Union[str, dict]: The generated text or an error dictionary.
        """
        # 1. Export the discussion to the lollms-native text format
        prompt_text = discussion.export("lollms_text", branch_tip_id)
        
        # 2. Extract images from the LAST message of the branch
        # lollms-webui's endpoint associates images with the final prompt
        active_branch_id = branch_tip_id or discussion.active_branch_id
        branch = discussion.get_branch(active_branch_id)
        last_message = branch[-1] if branch else None
        
        image_data = []
        if last_message and last_message.images:
            # The endpoint expects a list of base64 strings.
            # We will only process images of type 'base64'. URL types are not supported by this endpoint.
            for img in last_message.images:
                if img['type'] == 'base64':
                    image_data.append(img['data'])
                # Note: 'url' type images are ignored for this binding.

        # 3. Determine endpoint and build payload
        endpoint = "/lollms_generate_with_images" if image_data else "/lollms_generate"
        url = f"{self.host_address}{endpoint}"
        
        headers = {'Content-Type': 'application/json'}
        if self.service_key:
            headers['Authorization'] = f'Bearer {self.service_key}'

        data = {
            "prompt": prompt_text,
            "model_name": self.model_name,
            "personality": self.personality,
            "n_predict": n_predict,
            "stream": stream,
            "temperature": temperature,
            "top_k": top_k,
            "top_p": top_p,
            "repeat_penalty": repeat_penalty,
            "repeat_last_n": repeat_last_n,
            "seed": seed,
            "n_threads": n_threads
        }
        if image_data:
            data["images"] = image_data

        # 4. Make the request (logic copied and adapted from generate_text)
        try:
            response = requests.post(
                url, 
                json=data, 
                headers=headers, 
                stream=stream,
                verify=self.verify_ssl_certificate
            )
            response.raise_for_status() # Raise an exception for bad status codes

            if not stream:
                return response.text.strip()
            else:
                full_response_text = ""
                for line in response.iter_lines():
                    if line:
                        chunk = line.decode("utf-8")
                        full_response_text += chunk
                        if streaming_callback:
                            if not streaming_callback(chunk, MSG_TYPE.MSG_TYPE_CHUNK):
                                break
                # Clean up potential quotes from some streaming formats
                if full_response_text.startswith('"') and full_response_text.endswith('"'):
                    full_response_text = full_response_text[1:-1]
                return full_response_text.rstrip('!')

        except requests.exceptions.RequestException as e:
            error_message = f"lollms-webui request error: {e}"
            return {"status": "error", "message": error_message}
        except Exception as ex:
            error_message = f"lollms-webui generation error: {str(ex)}"
            return {"status": "error", "message": error_message}    
    def tokenize(self, text: str) -> list:
        """
        Tokenize the input text into a list of tokens using the /lollms_tokenize endpoint.

        Args:
            text (str): The text to tokenize.

        Returns:
            list: List of tokens.
        """
        response=None
        try:
            # Prepare the request payload
            payload = {
                "prompt": text,
                "return_named": False  # Set to True if you want named tokens
            }
            
            # Make the POST request to the /lollms_tokenize endpoint
            response = requests.post(f"{self.host_address}/lollms_tokenize", json=payload)
            
            # Check if the request was successful
            if response.status_code == 200:
                return response.json()
            else:
                raise Exception(f"Failed to tokenize text: {response.text}")
        except Exception as ex:
            trace_exception(ex)
            raise Exception(f"Failed to tokenize text: {response.text}")
          
    def detokenize(self, tokens: list) -> str:
        """
        Convert a list of tokens back to text using the /lollms_detokenize endpoint.

        Args:
            tokens (list): List of tokens to detokenize.

        Returns:
            str: Detokenized text.
        """
        try:
            # Prepare the request payload
            payload = {
                "tokens": tokens,
                "return_named": False  # Set to True if you want named tokens
            }

            # Make the POST request to the /lollms_detokenize endpoint
            response = requests.post(f"{self.host_address}/lollms_detokenize", json=payload)

            # Check if the request was successful
            if response.status_code == 200:
                return response.json()
            else:
                raise Exception(f"Failed to detokenize tokens: {response.text}")
        except Exception as ex:
            return {"status": False, "error": str(ex)}

    def count_tokens(self, text: str) -> int:
        """
        Count tokens from a text.

        Args:
            tokens (list): List of tokens to detokenize.

        Returns:
            int: Number of tokens in text.
        """        
        return len(self.tokenize(text))
    
    def embed(self, text: str, **kwargs) -> list:
        """
        Get embeddings for the input text using Ollama API
        
        Args:
            text (str or List[str]): Input text to embed
            **kwargs: Additional arguments like model, truncate, options, keep_alive
        
        Returns:
            dict: Response containing embeddings
        """
        api_key = kwargs.pop("api_key", None)
        headers = (
            {"Content-Type": "application/json", "Authorization": api_key}
            if api_key
            else {"Content-Type": "application/json"}
        )
        embeddings = []
        request_data = {"text": text}
        response = requests.post(f"{self.host_address}/lollms_embed", json=request_data, headers=headers)
        response.raise_for_status()
        result = response.json()
        return result["vector"]

    def get_model_info(self) -> dict:
        """
        Return information about the current LOLLMS model.

        Returns:
            dict: Dictionary containing model name, version, host address, and personality.
        """
        return {
            "name": "lollms",
            "version": "1.0",
            "host_address": self.host_address,
            "model_name": self.model_name,
            "personality": self.personality
        }


    def list_models(self) -> dict:
        """Lists models"""
        url = f"{self.host_address}/list_models"

        response = requests.get(url)

        if response.status_code == 200:
            try:
                models = json.loads(response.content.decode("utf-8"))
                return [{"model_name":m} for m in models]
            except Exception as ex:
                return {"status": False, "error": str(ex)}
        else:
            return {"status": False, "error": response.text}


    def load_model(self, model_name: str) -> bool:
        """
        Load a specific model into the LOLLMS binding.

        Args:
            model_name (str): Name of the model to load.

        Returns:
            bool: True if model loaded successfully.
        """
        self.model = model_name
        self.model_name = model_name
        return True

    # Lollms specific methods
    def lollms_listMountedPersonalities(self, host_address:str=None):
        host_address = host_address if host_address else self.host_address
        url = f"{host_address}/list_mounted_personalities"

        response = requests.get(url)

        if response.status_code == 200:
            try:
                text = json.loads(response.content.decode("utf-8"))
                return text
            except Exception as ex:
                return {"status": False, "error": str(ex)}
        else:
            return {"status": False, "error": response.text}



import os
import requests
from pathlib import Path
from typing import Optional, List, Dict, Any

from lollms_client.lollms_ttm_binding import LollmsTTMBinding
from ascii_colors import trace_exception, ASCIIColors
import pipmaster as pm

# Ensure required packages are installed
pm.ensure_packages(["requests"])

BindingName = "TopMediaiTTMBinding"

class TopMediaiTTMBinding(LollmsTTMBinding):
    """A Text-to-Music binding for the TopMediai API."""

    def __init__(self,
                 **kwargs):
        # Prioritize 'model_name' but accept 'model' as an alias from config files.
        if 'model' in kwargs and 'model_name' not in kwargs:
            kwargs['model_name'] = kwargs.pop('model')
        self.api_key = self.config.get("api_key") or os.environ.get("TOPMEDIAI_API_KEY")
        if not self.api_key:
            raise ValueError("TopMediai API key is required. Please set it in config or as TOPMEDIAI_API_KEY env var.")
        self.base_url = "https://api.topmediai.com/v1"
        self.headers = {"x-api-key": self.api_key, "Content-Type": "application/json"}

    def list_models(self, **kwargs) -> List[str]:
        # The API does not provide a list of selectable models.
        # It's a single, prompt-based system.
        return ["default"]

    def generate_music(self, prompt: str, **kwargs) -> bytes:
        """
        Generates music using the TopMediai synchronous API.
        """
        url = f"{self.base_url}/music"
        duration = kwargs.get("duration", 30)

        payload = {
            "text": prompt,
            "duration": f"{duration}", # API expects duration as a string
        }

        try:
            ASCIIColors.info("Requesting music from TopMediai...")
            response = requests.post(url, json=payload, headers=self.headers)
            response.raise_for_status()
            data = response.json()

            if data.get("code") != 0:
                raise Exception(f"TopMediai API returned an error: {data.get('message', 'Unknown error')}")

            audio_url = data.get("data", {}).get("music_url")
            if not audio_url:
                raise Exception("API response did not contain a music URL.")

            ASCIIColors.info(f"Downloading generated audio from {audio_url}")
            audio_response = requests.get(audio_url)
            audio_response.raise_for_status()
            
            return audio_response.content

        except requests.exceptions.HTTPError as e:
            try:
                error_details = e.response.json()
                raise Exception(f"TopMediai API HTTP Error: {error_details}") from e
            except:
                raise Exception(f"TopMediai API HTTP Error: {e.response.text}") from e
        except Exception as e:
            trace_exception(e)
            raise

if __name__ == '__main__':
    ASCIIColors.magenta("--- TopMediai TTM Binding Test ---")
    if "TOPMEDIAI_API_KEY" not in os.environ:
        ASCIIColors.error("TOPMEDIAI_API_KEY environment variable not set. Cannot run test.")
        exit(1)
        
    try:
        binding = TopMediaiTTMBinding()
        
        ASCIIColors.cyan("\n--- Test: Music Generation ---")
        prompt = "lo-fi hip hop beat, chill, relaxing, perfect for studying"
        music_bytes = binding.generate_music(prompt, duration=30)
        
        assert len(music_bytes) > 1000, "Generated music bytes are too small."
        output_path = Path(__file__).parent / "tmp_topmediai_music.mp3"
        with open(output_path, "wb") as f:
            f.write(music_bytes)
        ASCIIColors.green(f"Music generation OK. Audio saved to {output_path}")

    except Exception as e:
        trace_exception(e)
        ASCIIColors.error(f"TopMediai TTM binding test failed: {e}")


import os
import requests
import time
from pathlib import Path
from typing import Optional, List, Dict, Any

from lollms_client.lollms_ttm_binding import LollmsTTMBinding
from ascii_colors import trace_exception, ASCIIColors
import pipmaster as pm

# Ensure required packages are installed
pm.ensure_packages(["requests"])

BindingName = "ReplicateTTMBinding"

# Popular music models available on Replicate
# Sourced from: https://replicate.com/collections/text-to-music
REPLICATE_MODELS = [
    {"model_name": "meta/musicgen:b05b1dff1d8c6ac63d42422dd565e23b63869bf2d51acda751e04b5dd304535d", "display_name": "Meta - MusicGen", "description": "State-of-the-art controllable text-to-music model from Meta."},
    {"model_name": "suno-ai/bark:b76242b40d67c76ab6742e987628a2a9ac019e11d56ab96c4e91ce03b79b2787", "display_name": "Suno - Bark", "description": "Text-to-audio model capable of music, voice, and sound effects."},
    {"model_name": "joehoover/musicgen-melody:7a76a8258b23fae65c5a24debbe88414f9bed22c2422a63465731103f6990803", "display_name": "MusicGen Melody", "description": "MusicGen fine-tuned for generating melodies."},
]

class ReplicateTTMBinding(LollmsTTMBinding):
    """A Text-to-Music binding for models hosted on Replicate."""

    def __init__(self,
                 **kwargs):
        # Prioritize 'model_name' but accept 'model' as an alias from config files.
        if 'model' in kwargs and 'model_name' not in kwargs:
            kwargs['model_name'] = kwargs.pop('model')
        self.api_key = self.config.get("api_key") or os.environ.get("REPLICATE_API_TOKEN")
        if not self.api_key:
            raise ValueError("Replicate API token is required. Please set it in config or as REPLICATE_API_TOKEN env var.")
        self.model_version = self.config.get("model_name", "meta/musicgen:b05b1dff1d8c6ac63d42422dd565e23b63869bf2d51acda751e04b5dd304535d")
        self.base_url = "https://api.replicate.com/v1"
        self.headers = {"Authorization": f"Token {self.api_key}", "Content-Type": "application/json"}

    def list_models(self, **kwargs) -> List[Dict[str, str]]:
        return REPLICATE_MODELS

    def generate_music(self, prompt: str, **kwargs) -> bytes:
        """
        Generates music via Replicate by starting a prediction and polling for the result.
        """
        model_id, version_id = self.model_version.split(":")
        
        payload = {
            "version": version_id,
            "input": {
                "prompt": prompt,
                "duration": kwargs.get("duration", 8),
                "temperature": kwargs.get("temperature", 1.0),
                "top_p": kwargs.get("top_p", 0.9),
                # Add other model-specific parameters here
            }
        }
        
        try:
            # 1. Start the prediction
            ASCIIColors.info(f"Submitting music generation job to Replicate ({model_id})...")
            start_response = requests.post(f"{self.base_url}/predictions", json=payload, headers=self.headers)
            start_response.raise_for_status()
            job_data = start_response.json()
            get_url = job_data["urls"]["get"]
            ASCIIColors.info(f"Job submitted. Polling for results at: {get_url}")

            # 2. Poll for the result
            while True:
                poll_response = requests.get(get_url, headers=self.headers)
                poll_response.raise_for_status()
                poll_data = poll_response.json()
                status = poll_data["status"]

                if status == "succeeded":
                    ASCIIColors.green("Generation successful!")
                    output_url = poll_data["output"]
                    # Download the resulting audio file
                    audio_response = requests.get(output_url)
                    audio_response.raise_for_status()
                    return audio_response.content
                elif status in ["starting", "processing"]:
                    ASCIIColors.info(f"Job status: {status}. Waiting...")
                    time.sleep(3)
                else: # failed, canceled
                    error_log = poll_data.get("logs", "No logs available.")
                    raise Exception(f"Replicate job failed with status '{status}'. Log: {error_log}")
                    
        except requests.exceptions.HTTPError as e:
            error_details = e.response.json().get("detail", e.response.text)
            raise Exception(f"Replicate API HTTP Error: {error_details}") from e
        except Exception as e:
            trace_exception(e)
            raise

if __name__ == '__main__':
    ASCIIColors.magenta("--- Replicate TTM Binding Test ---")
    if "REPLICATE_API_TOKEN" not in os.environ:
        ASCIIColors.error("REPLICATE_API_TOKEN environment variable not set. Cannot run test.")
        exit(1)
        
    try:
        binding = ReplicateTTMBinding()
        
        ASCIIColors.cyan("\n--- Test: Music Generation with MusicGen ---")
        prompt = "An epic cinematic orchestral piece, with soaring strings and dramatic percussion, fit for a movie trailer"
        music_bytes = binding.generate_music(prompt, duration=10)
        
        assert len(music_bytes) > 1000, "Generated music bytes are too small."
        output_path = Path(__file__).parent / "tmp_replicate_music.wav"
        with open(output_path, "wb") as f:
            f.write(music_bytes)
        ASCIIColors.green(f"Music generation OK. Audio saved to {output_path}")

    except Exception as e:
        trace_exception(e)
        ASCIIColors.error(f"Replicate TTM binding test failed: {e}")


import os
import importlib
from io import BytesIO
from typing import Optional, List, Dict, Any, Union, Tuple
from pathlib import Path
import base64
import threading
import queue
from concurrent.futures import Future
import time
import hashlib
import requests
from tqdm import tqdm
import json
import shutil
import numpy as np
import gc
import argparse
import uvicorn
from fastapi import FastAPI, APIRouter, HTTPException, UploadFile, Form
from fastapi import Request, Response
from fastapi.responses import Response
from pydantic import BaseModel, Field
import sys
import platform
import inspect

class PullModelRequest(BaseModel):
    hf_id: Optional[str] = Field(default=None, description="Hugging Face repo id or URL, e.g. 'stabilityai/sdxl-turbo'")
    safetensors_url: Optional[str] = Field(default=None, description="Direct URL to a .safetensors file")
    local_name: Optional[str] = Field(default=None, description="Optional name/folder under models/")

# Add binding root to sys.path to ensure local modules can be imported if structured that way.
binding_root = Path(__file__).resolve().parent.parent
sys.path.insert(0, str(binding_root))

# --- Dependency Check and Imports ---
try:
    import torch
    from diffusers import (
        AutoPipelineForText2Image, AutoPipelineForImage2Image, AutoPipelineForInpainting,
        DiffusionPipeline, StableDiffusionPipeline, QwenImageEditPipeline, QwenImageEditPlusPipeline
    )
    from diffusers.utils import load_image
    from PIL import Image
    from ascii_colors import trace_exception, ASCIIColors
    DIFFUSERS_AVAILABLE = True
except ImportError as e:
    print(f"FATAL: A required package is missing from the server's venv: {e}.")
    DIFFUSERS_AVAILABLE = False
    # Define dummy classes to allow server to start and report error via API
    class Dummy: pass
    torch = Dummy()
    torch.cuda = Dummy()
    torch.cuda.is_available = lambda: False
    torch.backends = Dummy()
    torch.backends.mps = Dummy()
    torch.backends.mps.is_available = lambda: False
    AutoPipelineForText2Image = AutoPipelineForImage2Image = AutoPipelineForInpainting = DiffusionPipeline = StableDiffusionPipeline = QwenImageEditPipeline = QwenImageEditPlusPipeline = Image = load_image = ASCIIColors = trace_exception = Dummy

# --- Server Setup ---
app = FastAPI(title="Diffusers TTI Server")
router = APIRouter()
MODELS_PATH = Path("./models")

# --- START: Core Logic (Complete and Unabridged) ---
CIVITAI_MODELS = {
    "DreamShaper-8": {
        "display_name": "DreamShaper 8", "url": "https://civitai.com/api/download/models/128713",
        "filename": "dreamshaper_8.safetensors", "description": "Versatile SD1.5 style model.", "owned_by": "civitai"
    },
    "Juggernaut-xl": {
        "display_name": "Juggernaut XL", "url": "https://civitai.com/api/download/models/133005",
        "filename": "juggernautXL_version6Rundiffusion.safetensors", "description": "Artistic SDXL.", "owned_by": "civitai"
    },
}

HF_PUBLIC_MODELS = {
    "General Purpose & SDXL": [
        {"model_name": "stabilityai/stable-diffusion-xl-base-1.0", "display_name": "Stable Diffusion XL 1.0", "desc": "Official 1024x1024 text-to-image model from Stability AI."},
        {"model_name": "stabilityai/sdxl-turbo", "display_name": "SDXL Turbo", "desc": "A fast, real-time text-to-image model based on SDXL."},
        {"model_name": "kandinsky-community/kandinsky-3", "display_name": "Kandinsky 3", "desc": "A powerful multilingual model with strong prompt understanding and aesthetic quality."},
        {"model_name": "playgroundai/playground-v2.5-1024px-aesthetic", "display_name": "Playground v2.5", "desc": "A high-quality model focused on aesthetic outputs."},
    ],
    "Photorealistic": [
        {"model_name": "emilianJR/epiCRealism", "display_name": "epiCRealism", "desc": "A popular community model for generating photorealistic images."},
        {"model_name": "SG161222/Realistic_Vision_V5.1_noVAE", "display_name": "Realistic Vision 5.1", "desc": "One of the most popular realistic models, great for portraits and scenes."},
        {"model_name": "Photon-v1", "display_name": "Photon", "desc": "A model known for high-quality, realistic images with good lighting and detail."},
    ],
    "Anime & Illustration": [
        {"model_name": "hakurei/waifu-diffusion", "display_name": "Waifu Diffusion 1.4", "desc": "A widely-used model for generating high-quality anime-style images."},
        {"model_name": "gsdf/Counterfeit-V3.0", "display_name": "Counterfeit V3.0", "desc": "A strong model for illustrative and 2.5D anime styles."},
        {"model_name": "cagliostrolab/animagine-xl-3.0", "display_name": "Animagine XL 3.0", "desc": "A state-of-the-art anime model on the SDXL architecture."},
    ],
    "Artistic & Stylized": [
        {"model_name": "wavymulder/Analog-Diffusion", "display_name": "Analog Diffusion", "desc": "Creates images with a vintage, analog film aesthetic."},
        {"model_name": "dreamlike-art/dreamlike-photoreal-2.0", "display_name": "Dreamlike Photoreal 2.0", "desc": "Produces stunning, artistic, and photorealistic images."},
    ],
    "Image Editing Tools": [
        {"model_name": "stabilityai/stable-diffusion-xl-refiner-1.0", "display_name": "SDXL Refiner 1.0", "desc": "A dedicated refiner model to improve details in SDXL generations."},
        {"model_name": "timbrooks/instruct-pix2pix", "display_name": "Instruct-Pix2Pix", "desc": "The original instruction-based image editing model (SD 1.5)."},
        {"model_name": "kandinsky-community/kandinsky-2-2-instruct-pix2pix", "display_name": "Kandinsky 2.2 Instruct", "desc": "An instruction-based model with strong prompt adherence, based on Kandinsky 2.2."},
        {"model_name": "diffusers/stable-diffusion-xl-1.0-inpainting-0.1", "display_name": "SDXL Inpainting", "desc": "A dedicated inpainting model based on SDXL 1.0 for filling in masked areas."},
        {"model_name": "Qwen/Qwen-Image-Edit", "display_name": "Qwen Image Edit", "desc": "An instruction-based model for various image editing tasks. (Review License)."},
        {"model_name": "Qwen/Qwen-Image-Edit-2509", "display_name": "Qwen Image Edit Plus", "desc": "Advanced multi-image editing and fusion. (Review License)."},
    ],
    "Legacy & Base Models": [
        {"model_name": "runwayml/stable-diffusion-v1-5", "display_name": "Stable Diffusion 1.5", "desc": "The classic and versatile SD1.5 base model."},
        {"model_name": "stabilityai/stable-diffusion-2-1", "display_name": "Stable Diffusion 2.1", "desc": "The 768x768 base model from the SD2.x series."},
    ]
}

HF_GATED_MODELS = {
    "Next-Generation (Gated Access Required)": [
        {"model_name": "stabilityai/stable-diffusion-3-medium-diffusers", "display_name": "Stable Diffusion 3 Medium", "desc": "State-of-the-art model with advanced prompt understanding. Requires free registration."},
        {"model_name": "black-forest-labs/FLUX.1-schnell", "display_name": "FLUX.1 Schnell", "desc": "A powerful and extremely fast next-generation model. Requires access request."},
        {"model_name": "black-forest-labs/FLUX.1-dev", "display_name": "FLUX.1 Dev", "desc": "The larger developer version of the FLUX.1 model. Requires access request."},
    ]
}


TORCH_DTYPE_MAP_STR_TO_OBJ = {
    "float16": getattr(torch, 'float16', 'float16'), "bfloat16": getattr(torch, 'bfloat16', 'bfloat16'),
    "float32": getattr(torch, 'float32', 'float32'), "auto": "auto"
}

SCHEDULER_MAPPING = {
    "default": None, "ddim": "DDIMScheduler", "ddpm": "DDPMScheduler", "deis_multistep": "DEISMultistepScheduler",
    "dpm_multistep": "DPMSolverMultistepScheduler", "dpm_multistep_karras": "DPMSolverMultistepScheduler", "dpm_single": "DPMSolverSinglestepScheduler",
    "dpm_adaptive": "DPMSolverPlusPlusScheduler", "dpm++_2m": "DPMSolverMultistepScheduler", "dpm++_2m_karras": "DPMSolverMultistepScheduler",
    "dpm++_2s_ancestral": "DPMSolverAncestralDiscreteScheduler", "dpm++_2s_ancestral_karras": "DPMSolverAncestralDiscreteScheduler", "dpm++_sde": "DPMSolverSDEScheduler",
    "dpm++_sde_karras": "DPMSolverSDEScheduler", "euler_ancestral_discrete": "EulerAncestralDiscreteScheduler", "euler_discrete": "EulerDiscreteScheduler",
    "heun_discrete": "HeunDiscreteScheduler", "heun_karras": "HeunDiscreteScheduler", "lms_discrete": "LMSDiscreteScheduler",
    "lms_karras": "LMSDiscreteScheduler", "pndm": "PNDMScheduler", "unipc_multistep": "UniPCMultistepScheduler",
    "dpm++_2m_sde": "DPMSolverMultistepScheduler", "dpm++_2m_sde_karras": "DPMSolverMultistepScheduler", "dpm2": "KDPM2DiscreteScheduler",
    "dpm2_karras": "KDPM2DiscreteScheduler", "dpm2_a": "KDPM2AncestralDiscreteScheduler", "dpm2_a_karras": "KDPM2AncestralDiscreteScheduler",
    "euler": "EulerDiscreteScheduler", "euler_a": "EulerAncestralDiscreteScheduler", "heun": "HeunDiscreteScheduler", "lms": "LMSDiscreteScheduler"
}

SCHEDULER_USES_KARRAS_SIGMAS = [
    "dpm_multistep_karras","dpm++_2m_karras","dpm++_2s_ancestral_karras", "dpm++_sde_karras","heun_karras","lms_karras",
    "dpm++_2m_sde_karras","dpm2_karras","dpm2_a_karras"
]


class ModelManager:
    def __init__(self, config: Dict[str, Any], models_path: Path, registry: 'PipelineRegistry'):
        self.config = config
        self.models_path = models_path
        self.registry = registry
        self.pipeline: Optional[DiffusionPipeline] = None
        self.current_task: Optional[str] = None
        self.ref_count = 0
        self.lock = threading.Lock()
        self.queue = queue.Queue()
        self.is_loaded = False
        self.last_used_time = time.time()
        self._stop_event = threading.Event()
        self.worker_thread = threading.Thread(target=self._generation_worker, daemon=True)
        self.worker_thread.start()
        self._stop_monitor_event = threading.Event()
        self._unload_monitor_thread = None
        self._start_unload_monitor()
        self.supported_args: Optional[set] = None

    def acquire(self):
        with self.lock:
            self.ref_count += 1
            return self

    def release(self):
        with self.lock:
            self.ref_count -= 1
            return self.ref_count

    def stop(self):
        self._stop_event.set()
        if self._unload_monitor_thread:
            self._stop_monitor_event.set()
            self._unload_monitor_thread.join(timeout=2)
        self.queue.put(None)
        self.worker_thread.join(timeout=5)

    def _start_unload_monitor(self):
        unload_after = self.config.get("unload_inactive_model_after", 0)
        if unload_after > 0 and self._unload_monitor_thread is None:
            self._stop_monitor_event.clear()
            self._unload_monitor_thread = threading.Thread(target=self._unload_monitor, daemon=True)
            self._unload_monitor_thread.start()

    def _unload_monitor(self):
        unload_after = self.config.get("unload_inactive_model_after", 0)
        if unload_after <= 0:
            return
        ASCIIColors.info(f"Starting inactivity monitor for '{self.config['model_name']}' (timeout: {unload_after}s).")
        while not self._stop_monitor_event.wait(timeout=5.0):
            with self.lock:
                if not self.is_loaded:
                    continue
                if time.time() - self.last_used_time > unload_after:
                    ASCIIColors.info(f"Model '{self.config['model_name']}' has been inactive. Unloading.")
                    self._unload_pipeline()

    def _resolve_model_path(self, model_name: str) -> Union[str, Path]:
        path_obj = Path(model_name)
        if path_obj.is_absolute() and path_obj.exists():
            return model_name
        if model_name in CIVITAI_MODELS:
            filename = CIVITAI_MODELS[model_name]["filename"]
            local_path = self.models_path / filename
            if not local_path.exists():
                self._download_civitai_model(model_name)
            return local_path

        # Search in extra models path
        if state.extra_models_path and state.extra_models_path.exists():
            found_paths = list(state.extra_models_path.rglob(model_name))
            if found_paths:
                ASCIIColors.info(f"Found model in extra path: {found_paths[0]}")
                return found_paths[0]

        # Search in primary models path
        found_paths = list(self.models_path.rglob(model_name))
        if found_paths:
            ASCIIColors.info(f"Found model in primary path: {found_paths[0]}")
            return found_paths[0]

        # Fallback for HF hub models that are folders, not single files.
        local_path = self.models_path / model_name
        if local_path.exists():
            return local_path

        return model_name

    def _download_civitai_model(self, model_key: str):
        model_info = CIVITAI_MODELS[model_key]
        url = model_info["url"]
        filename = model_info["filename"]
        dest_path = self.models_path / filename
        temp_path = dest_path.with_suffix(".temp")
        ASCIIColors.cyan(f"Downloading '{filename}' from Civitai... to {dest_path}")
        try:
            with requests.get(url, stream=True) as r:
                r.raise_for_status()
                total_size = int(r.headers.get('content-length', 0))
                with open(temp_path, 'wb') as f, tqdm(total=total_size, unit='iB', unit_scale=True, desc=f"Downloading {filename}") as bar:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)
                        bar.update(len(chunk))
            shutil.move(temp_path, dest_path)
            ASCIIColors.green(f"Model '{filename}' downloaded successfully.")
        except Exception as e:
            if temp_path.exists():
                temp_path.unlink()
            raise Exception(f"Failed to download model {filename}: {e}")

    def _set_scheduler(self):
        if not self.pipeline:
            return
        if "Qwen" in self.config.get("model_name", "") or "FLUX" in self.config.get("model_name", ""):
            ASCIIColors.info("Special model detected, skipping custom scheduler setup.")
            return
        scheduler_name_key = self.config["scheduler_name"].lower()
        if scheduler_name_key == "default":
            return
        scheduler_class_name = SCHEDULER_MAPPING.get(scheduler_name_key)
        if scheduler_class_name:
            try:
                SchedulerClass = getattr(importlib.import_module("diffusers.schedulers"), scheduler_class_name)
                scheduler_config = self.pipeline.scheduler.config
                scheduler_config["use_karras_sigmas"] = scheduler_name_key in SCHEDULER_USES_KARRAS_SIGMAS
                self.pipeline.scheduler = SchedulerClass.from_config(scheduler_config)
                ASCIIColors.info(f"Switched scheduler to {scheduler_class_name}")
            except Exception as e:
                ASCIIColors.warning(f"Could not switch scheduler to {scheduler_name_key}: {e}. Using current default.")

    def _execute_load_pipeline(self, task: str, model_path: Union[str, Path], torch_dtype: Any):
        if platform.system() == "Windows":
            os.environ["HF_HUB_ENABLE_SYMLINKS"] = "0"
        
        model_name_from_config = self.config.get("model_name", "")
        use_device_map = False

        try:
            load_params = {}
            if self.config.get("hf_cache_path"):
                load_params["cache_dir"] = str(self.config["hf_cache_path"])
            load_params["torch_dtype"] = torch_dtype

            is_qwen_model = "Qwen" in model_name_from_config
            is_flux_model = "FLUX" in model_name_from_config

            if is_qwen_model or is_flux_model:
                ASCIIColors.info(f"Special model '{model_name_from_config}' detected. Using dedicated pipeline loader.")
                load_params.update({
                    "use_safetensors": self.config["use_safetensors"],
                    "token": self.config["hf_token"],
                    "local_files_only": self.config["local_files_only"]
                })
                if self.config["hf_variant"]:
                    load_params["variant"] = self.config["hf_variant"]
                if not self.config["safety_checker_on"]:
                    load_params["safety_checker"] = None
                
                should_offload = self.config["enable_cpu_offload"] or self.config["enable_sequential_cpu_offload"]
                if should_offload:
                    ASCIIColors.info(f"Offload enabled. Forcing device_map='auto' for {model_name_from_config}.")
                    use_device_map = True
                    load_params["device_map"] = "auto"
                
                if is_flux_model:
                    self.pipeline = AutoPipelineForText2Image.from_pretrained(model_name_from_config, **load_params)
                elif "Qwen-Image-Edit-2509" in model_name_from_config:
                    self.pipeline = QwenImageEditPlusPipeline.from_pretrained(model_name_from_config, **load_params)
                elif "Qwen-Image-Edit" in model_name_from_config:
                    self.pipeline = QwenImageEditPipeline.from_pretrained(model_name_from_config, **load_params)
                elif "Qwen/Qwen-Image" in model_name_from_config:
                    self.pipeline = DiffusionPipeline.from_pretrained(model_name_from_config, **load_params)
            
            else:
                is_safetensors_file = str(model_path).endswith(".safetensors")
                if is_safetensors_file:
                    ASCIIColors.info(f"Loading standard model from local .safetensors file: {model_path}")
                    try:
                        self.pipeline = AutoPipelineForText2Image.from_single_file(model_path, **load_params)
                    except Exception as e:
                        ASCIIColors.warning(f"Failed to load with AutoPipeline, falling back to StableDiffusionPipeline: {e}")
                        self.pipeline = StableDiffusionPipeline.from_single_file(model_path, **load_params)
                else:
                    ASCIIColors.info(f"Loading standard model from Hub: {model_path}")
                    load_params.update({
                        "use_safetensors": self.config["use_safetensors"],
                        "token": self.config["hf_token"],
                        "local_files_only": self.config["local_files_only"]
                    })
                    if self.config["hf_variant"]:
                        load_params["variant"] = self.config["hf_variant"]
                    if not self.config["safety_checker_on"]:
                        load_params["safety_checker"] = None
                    
                    is_large_model = "stable-diffusion-3" in str(model_path)
                    should_offload = self.config["enable_cpu_offload"] or self.config["enable_sequential_cpu_offload"]
                    if is_large_model and should_offload:
                        ASCIIColors.info(f"Large model '{model_path}' detected with offload enabled. Using device_map='auto'.")
                        use_device_map = True
                        load_params["device_map"] = "auto"

                    if task == "text2image":
                        self.pipeline = AutoPipelineForText2Image.from_pretrained(model_path, **load_params)
                    elif task == "image2image":
                        self.pipeline = AutoPipelineForImage2Image.from_pretrained(model_path, **load_params)
                    elif task == "inpainting":
                        self.pipeline = AutoPipelineForInpainting.from_pretrained(model_path, **load_params)
        
        except Exception as e:
            error_str = str(e).lower()
            if "401" in error_str or "gated" in error_str or "authorization" in error_str:
                msg = (f"AUTHENTICATION FAILED for model '{model_name_from_config}'. Please ensure you accepted the model license and provided a valid HF token.")
                raise RuntimeError(msg) 
            raise e

        self._set_scheduler()

        if not use_device_map:
            self.pipeline.to(self.config["device"])
            if self.config["enable_xformers"]:
                try:
                    self.pipeline.enable_xformers_memory_efficient_attention()
                except Exception as e:
                    ASCIIColors.warning(f"Could not enable xFormers: {e}.")
            
            if self.config["enable_cpu_offload"] and self.config["device"] != "cpu":
                self.pipeline.enable_model_cpu_offload()
            elif self.config["enable_sequential_cpu_offload"] and self.config["device"] != "cpu":
                self.pipeline.enable_sequential_cpu_offload()
        else:
             ASCIIColors.info("Device map handled device placement. Skipping manual pipeline.to() and offload calls.")

        if self.pipeline:
            sig = inspect.signature(self.pipeline.__call__)
            self.supported_args = {p.name for p in sig.parameters.values()}
            ASCIIColors.info(f"Pipeline supported arguments detected: {self.supported_args}")

        self.is_loaded = True
        self.current_task = task
        self.last_used_time = time.time()
        ASCIIColors.green(f"Model '{model_name_from_config}' loaded successfully using '{'device_map' if use_device_map else 'standard'}' mode for task '{task}'.")

    def _load_pipeline_for_task(self, task: str):
        if self.pipeline and self.current_task == task:
            return
        if self.pipeline:
            self._unload_pipeline()
        
        model_name = self.config.get("model_name", "")
        if not model_name:
            raise ValueError("Model name cannot be empty for loading.")
        
        ASCIIColors.info(f"Loading Diffusers model: {model_name} for task: {task}")
        model_path = self._resolve_model_path(model_name)
        torch_dtype = TORCH_DTYPE_MAP_STR_TO_OBJ.get(self.config["torch_dtype_str"].lower())
        
        try:
            self._execute_load_pipeline(task, model_path, torch_dtype)
            return
        except Exception as e:
            is_oom = "out of memory" in str(e).lower()
            if not is_oom or not hasattr(self, 'registry'):
                raise e
        
        ASCIIColors.warning(f"Failed to load '{model_name}' due to OOM. Attempting to unload other models to free VRAM.")
        
        candidates_to_unload = [m for m in self.registry.get_all_managers() if m is not self and m.is_loaded]
        candidates_to_unload.sort(key=lambda m: m.last_used_time)

        if not candidates_to_unload:
            ASCIIColors.error("OOM error, but no other models are available to unload.")
            raise Exception("OOM error, but no other models are available to unload.")

        for victim in candidates_to_unload:
            ASCIIColors.info(f"Unloading '{victim.config['model_name']}' (last used: {time.ctime(victim.last_used_time)}) to free VRAM.")
            victim._unload_pipeline()
            
            try:
                ASCIIColors.info(f"Retrying to load '{model_name}'...")
                self._execute_load_pipeline(task, model_path, torch_dtype)
                ASCIIColors.green(f"Successfully loaded '{model_name}' after freeing VRAM.")
                return
            except Exception as retry_e:
                is_oom_retry = "out of memory" in str(retry_e).lower()
                if not is_oom_retry:
                    raise retry_e 
        
        ASCIIColors.error(f"Could not load '{model_name}' even after unloading all other models.")
        raise e

    def _unload_pipeline(self):
        if self.pipeline:
            model_name = self.config.get('model_name', 'Unknown')
            del self.pipeline
            self.pipeline = None
            self.supported_args = None
            gc.collect()
            if torch and torch.cuda.is_available():
                torch.cuda.empty_cache()
            self.is_loaded = False
            self.current_task = None
            ASCIIColors.info(f"Model '{model_name}' unloaded and VRAM cleared.")

    def _generation_worker(self):
        while not self._stop_event.is_set():
            try:
                job = self.queue.get(timeout=1)
                if job is None:
                    break
                future, task, pipeline_args = job
                output = None
                try:
                    with self.lock:
                        self.last_used_time = time.time()
                        if not self.is_loaded or self.current_task != task:
                            self._load_pipeline_for_task(task)
                    
                    if self.supported_args:
                        filtered_args = {k: v for k, v in pipeline_args.items() if k in self.supported_args}
                    else:
                        ASCIIColors.warning("Supported argument set not found. Using unfiltered arguments.")
                        filtered_args = pipeline_args

                    with torch.no_grad():
                        output = self.pipeline(**filtered_args)
                    
                    pil = output.images[0]
                    buf = BytesIO()
                    pil.save(buf, format="PNG")
                    future.set_result(buf.getvalue())
                except Exception as e:
                    trace_exception(e)
                    future.set_exception(e)
                finally:
                    self.queue.task_done()
                    if output is not None:
                        del output
                    gc.collect()
                    if torch.cuda.is_available():
                        torch.cuda.empty_cache()
            except queue.Empty:
                continue

class PipelineRegistry:
    _instance = None
    _lock = threading.Lock()
    def __new__(cls, *args, **kwargs):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super().__new__(cls)
                cls._instance._managers = {}
                cls._instance._registry_lock = threading.Lock()
        return cls._instance
    @staticmethod
    def _get_critical_keys():
        return [
            "model_name","device","torch_dtype_str","use_safetensors",
            "safety_checker_on","hf_variant","enable_cpu_offload",
            "enable_sequential_cpu_offload","enable_xformers",
            "local_files_only","hf_cache_path","unload_inactive_model_after"
        ]
    def _get_config_key(self, config: Dict[str, Any]) -> str:
        key_data = tuple(sorted((k, config.get(k)) for k in self._get_critical_keys()))
        return hashlib.sha256(str(key_data).encode('utf-8')).hexdigest()
    def get_manager(self, config: Dict[str, Any], models_path: Path) -> ModelManager:
        key = self._get_config_key(config)
        with self._registry_lock:
            if key not in self._managers:
                self._managers[key] = ModelManager(config.copy(), models_path, self)
            return self._managers[key].acquire()
    def release_manager(self, config: Dict[str, Any]):
        key = self._get_config_key(config)
        with self._registry_lock:
            if key in self._managers:
                manager = self._managers[key]
                ref_count = manager.release()
                if ref_count == 0:
                    ASCIIColors.info(f"Reference count for model '{config.get('model_name')}' is zero. Cleaning up manager.")
                    manager.stop()
                    with manager.lock:
                        manager._unload_pipeline()
                    del self._managers[key]
    def get_active_managers(self) -> List[ModelManager]:
        with self._registry_lock:
            return [m for m in self._managers.values() if m.is_loaded]
    def get_all_managers(self) -> List[ModelManager]:
        with self._registry_lock:
            return list(self._managers.values())

class ServerState:
    def __init__(self, models_path: Path, extra_models_path: Optional[Path] = None):
        self.models_path = models_path
        self.extra_models_path = extra_models_path
        self.models_path.mkdir(parents=True, exist_ok=True)
        if self.extra_models_path:
            self.extra_models_path.mkdir(parents=True, exist_ok=True)
        self.config_path = self.models_path.parent / "diffusers_server_config.json"
        self.registry = PipelineRegistry()
        self.manager: Optional[ModelManager] = None
        self.config = {}
        self.load_config()
        self._resolve_device_and_dtype()
        if self.config.get("model_name"):
            try:
                ASCIIColors.info(f"Acquiring initial model manager for '{self.config['model_name']}' on startup.")
                self.manager = self.registry.get_manager(self.config, self.models_path)
            except Exception as e:
                ASCIIColors.error(f"Failed to acquire model manager on startup: {e}")
                self.manager = None

    def get_default_config(self) -> Dict[str, Any]:
        return {
            "model_name": "", "device": "auto", "torch_dtype_str": "auto", "use_safetensors": True,
            "scheduler_name": "default", "safety_checker_on": True, "num_inference_steps": 25,
            "guidance_scale": 7.0, "width": 1024, "height": 1024, "seed": -1,
            "enable_cpu_offload": False, "enable_sequential_cpu_offload": False, "enable_xformers": False,
            "hf_variant": None, "hf_token": None, "hf_cache_path": None, "local_files_only": False,
            "unload_inactive_model_after": 0
        }

    def save_config(self):
        try:
            with open(self.config_path, 'w') as f:
                json.dump(self.config, f, indent=4)
            ASCIIColors.info(f"Server config saved to {self.config_path}")
        except Exception as e:
            ASCIIColors.error(f"Failed to save server config: {e}")

    def load_config(self):
        default_config = self.get_default_config()
        if self.config_path.exists():
            try:
                with open(self.config_path, 'r') as f:
                    loaded_config = json.load(f)
                default_config.update(loaded_config)
                self.config = default_config
                ASCIIColors.info(f"Loaded server configuration from {self.config_path}")
            except (json.JSONDecodeError, IOError) as e:
                ASCIIColors.warning(f"Could not load config file, using defaults. Error: {e}")
                self.config = default_config
        else:
            self.config = default_config
        self.save_config()

    def _resolve_device_and_dtype(self):
        if self.config.get("device", "auto").lower() == "auto":
            self.config["device"] = "cuda" if torch.cuda.is_available() else "mps" if torch.backends.mps.is_available() else "cpu"
        
        if ("Qwen" in self.config.get("model_name", "") or "FLUX" in self.config.get("model_name", "")) and self.config["device"] == "cuda":
            if hasattr(torch.cuda, 'is_bf16_supported') and torch.cuda.is_bf16_supported():
                self.config["torch_dtype_str"] = "bfloat16"
                ASCIIColors.info("Special model detected on compatible hardware. Forcing dtype to bfloat16 for stability.")
                return

        if self.config["torch_dtype_str"].lower() == "auto":
            self.config["torch_dtype_str"] = "float16" if self.config["device"] != "cpu" else "float32"

    def update_settings(self, new_settings: Dict[str, Any]):
        if 'model' in new_settings and 'model_name' not in new_settings:
            new_settings['model_name'] = new_settings.pop('model')

        if self.config.get("model_name") and not new_settings.get("model_name"):
            ASCIIColors.info("Incoming settings have no model_name. Preserving existing model.")
            new_settings["model_name"] = self.config["model_name"]

        if self.manager:
            self.registry.release_manager(self.manager.config)
            self.manager = None

        self.config.update(new_settings)
        ASCIIColors.info(f"Server config updated. Current model_name: {self.config.get('model_name')}")
        
        self._resolve_device_and_dtype()

        if self.config.get("model_name"):
            ASCIIColors.info("Acquiring model manager with updated configuration...")
            self.manager = self.registry.get_manager(self.config, self.models_path)
        else:
            ASCIIColors.warning("No model_name in config after update, manager not acquired.")
        
        self.save_config()
        return True

    def get_active_manager(self) -> ModelManager:
        if self.manager:
            return self.manager
        raise HTTPException(status_code=400, detail="No model is configured or manager is not active. Please set a model using the /set_settings endpoint.")

state: Optional[ServerState] = None

# --- Pydantic Models for API ---
class T2IRequest(BaseModel):
    prompt: str
    negative_prompt: str = ""
    params: Dict[str, Any] = Field(default_factory=dict)

class EditRequestPayload(BaseModel):
    prompt: str
    image_paths: List[str] = Field(default_factory=list)
    params: Dict[str, Any] = Field(default_factory=dict)

class EditRequestJSON(BaseModel):
    prompt: str
    images_b64: List[str] = Field(description="A list of Base64 encoded image strings.")
    params: Dict[str, Any] = Field(default_factory=dict)
def get_sanitized_request_for_logging(request_data: Any) -> Dict[str, Any]:
    """
    Takes a request object (Pydantic model or dict) and returns a 'safe' dictionary
    for logging, with long base64 strings replaced by placeholders.
    """
    import copy

    try:
        if hasattr(request_data, 'model_dump'):
            data = request_data.model_dump()
        elif isinstance(request_data, dict):
            data = copy.deepcopy(request_data)
        else:
            return {"error": "Unsupported data type for sanitization"}

        # Sanitize the main list of images
        if 'images_b64' in data and isinstance(data['images_b64'], list):
            count = len(data['images_b64'])
            data['images_b64'] = f"[<{count} base64 image(s) truncated>]"

        # Sanitize a potential mask in the 'params' dictionary
        if 'params' in data and isinstance(data.get('params'), dict):
            if 'mask_image' in data['params'] and isinstance(data['params']['mask_image'], str):
                original_len = len(data['params']['mask_image'])
                data['params']['mask_image'] = f"[<base64 mask truncated, len={original_len}>]"
        
        return data
    except Exception:
        return {"error": "Failed to sanitize request data."}

# --- API Endpoints ---
@router.post("/generate_image")
async def generate_image(request: T2IRequest):
    manager = None
    temp_config = None
    try:
        # Determine which model manager to use for this specific request
        if "model_name" in request.params and request.params["model_name"]:
            temp_config = state.config.copy()
            temp_config["model_name"] = request.params.pop("model_name") # Remove from params to avoid being passed to pipeline
            manager = state.registry.get_manager(temp_config, state.models_path)
            ASCIIColors.info(f"Using per-request model: {temp_config['model_name']}")
        else:
            manager = state.get_active_manager()
            ASCIIColors.info(f"Using session-configured model: {manager.config.get('model_name')}")

        # Start with the manager's config (base settings)
        pipeline_args = manager.config.copy()
        # Override with per-request parameters
        pipeline_args.update(request.params)

        # Add prompts and ensure types for specific args
        pipeline_args["prompt"] = request.prompt
        pipeline_args["negative_prompt"] = request.negative_prompt
        width = pipeline_args.get("width", 1024)
        height = pipeline_args.get("height", 1024)
        num_inference_steps = pipeline_args.get("num_inference_steps", 25)
        seed = pipeline_args.get("seed", -1)
        guidance_scale = pipeline_args.get("guidance_scale", 7.0)
        pipeline_args["width"] = int(width if width else 1024)
        pipeline_args["height"] = int(height if height else 1024)
        pipeline_args["num_inference_steps"] = int(num_inference_steps if num_inference_steps else 25)
        pipeline_args["guidance_scale"] = float(guidance_scale if guidance_scale else 7.0)

        seed = int(seed if seed is not None else -1)
        pipeline_args["generator"] = None
        if seed != -1:
            pipeline_args["generator"] = torch.Generator(device=manager.config["device"]).manual_seed(seed)
        
        model_name = manager.config.get("model_name", "")
        task = "text2image"
        
        if "Qwen-Image-Edit" in model_name:
            rng_seed = seed if seed != -1 else None
            rng = np.random.default_rng(seed=rng_seed)
            random_pixels = rng.integers(0, 256, size=(pipeline_args["height"], pipeline_args["width"], 3), dtype=np.uint8)
            placeholder_image = Image.fromarray(random_pixels, 'RGB')
            pipeline_args["image"] = placeholder_image
            pipeline_args["strength"] = float(pipeline_args.get("strength", 1.0))
            task = "image2image" 
        
        log_args = {k: v for k, v in pipeline_args.items() if k not in ['generator', 'image']}
        if pipeline_args.get("generator"): log_args['generator'] = f"<torch.Generator(seed={seed})>"
        if pipeline_args.get("image"): log_args['image'] = "<PIL Image object>"
        
        ASCIIColors.cyan("--- Generating Image with Settings ---")
        try:
            print(json.dumps(log_args, indent=2, default=str))
        except Exception as e:
            ASCIIColors.warning(f"Could not print all settings: {e}")
            print(log_args)
        ASCIIColors.cyan("------------------------------------")
        
        future = Future()
        manager.queue.put((future, task, pipeline_args))
        result_bytes = future.result()
        return Response(content=result_bytes, media_type="image/png")
    except Exception as e:
        trace_exception(e)
        raise HTTPException(status_code=500, detail=str(e))
    finally:
        if temp_config and manager:
            state.registry.release_manager(temp_config)
            ASCIIColors.info(f"Released per-request model: {temp_config['model_name']}")


@router.post("/edit_image")
async def edit_image(request: EditRequestJSON):
    manager = None
    temp_config = None
    ASCIIColors.info(f"Received /edit_image request with {len(request.images_b64)} image(s).")
    ASCIIColors.info(request.params)
    try:
        if "model_name" in request.params and request.params["model_name"]:
            temp_config = state.config.copy()
            temp_config["model_name"] = request.params.pop("model_name")
            manager = state.registry.get_manager(temp_config, state.models_path)
            ASCIIColors.info(f"Using per-request model: {temp_config['model_name']}")
        else:
            manager = state.get_active_manager()
            ASCIIColors.info(f"Using session-configured model: {manager.config.get('model_name')}")

        # Start with manager's config, then override with request params
        pipeline_args = manager.config.copy()
        pipeline_args.update(request.params)
        
        pipeline_args["prompt"] = request.prompt
        model_name = manager.config.get("model_name", "")

        pil_images = []
        for b64_string in request.images_b64:
            b64_data = b64_string.split(";base64,")[1] if ";base64," in b64_string else b64_string
            image_bytes = base64.b64decode(b64_data)
            pil_images.append(Image.open(BytesIO(image_bytes)).convert("RGB"))

        if not pil_images: raise HTTPException(status_code=400, detail="No valid images provided.")

        seed = int(pipeline_args.get("seed", -1))
        pipeline_args["generator"] = None
        if seed != -1: pipeline_args["generator"] = torch.Generator(device=manager.config["device"]).manual_seed(seed)
        
        if "mask_image" in pipeline_args and pipeline_args["mask_image"]:
            b64_mask = pipeline_args["mask_image"]
            b64_data = b64_mask.split(";base64,")[1] if ";base64," in b64_mask else b64_mask
            mask_bytes = base64.b64decode(b64_data)
            pipeline_args["mask_image"] = Image.open(BytesIO(mask_bytes)).convert("L")
        
        task = "inpainting" if "mask_image" in pipeline_args and pipeline_args["mask_image"] else "image2image"
        
        if "Qwen-Image-Edit-2509" in model_name:
            task = "image2image"
            pipeline_args.update({"true_cfg_scale": 4.0, "guidance_scale": 1.0, "num_inference_steps": 40, "negative_prompt": " "})
            edit_mode = pipeline_args.get("edit_mode", "fusion")
            if edit_mode == "fusion": pipeline_args["image"] = pil_images
        else:
            pipeline_args.update({"image": pil_images[0]})

        log_args = {k: v for k, v in pipeline_args.items() if k not in ['generator', 'image', 'mask_image']}
        if pipeline_args.get("generator"): log_args['generator'] = f"<torch.Generator(seed={seed})>"
        if 'image' in pipeline_args: log_args['image'] = f"[<{len(pil_images)} PIL Image(s)>]"
        if 'mask_image' in pipeline_args and pipeline_args['mask_image']: log_args['mask_image'] = "<PIL Mask Image>"

        ASCIIColors.cyan("--- Editing Image with Settings ---")
        try:
            print(json.dumps(log_args, indent=2, default=str))
        except Exception as e:
            ASCIIColors.warning(f"Could not print all settings: {e}")
            print(log_args)
        ASCIIColors.cyan("---------------------------------")
        
        future = Future(); manager.queue.put((future, task, pipeline_args))
        return Response(content=future.result(), media_type="image/png")
    except Exception as e:
        sanitized_payload = get_sanitized_request_for_logging(request)
        ASCIIColors.error(f"Exception in /edit_image. Sanitized Payload: {json.dumps(sanitized_payload, indent=2)}")
        trace_exception(e)
        raise HTTPException(status_code=500, detail=str(e))
    finally:
        if temp_config and manager:
            state.registry.release_manager(temp_config)
            ASCIIColors.info(f"Released per-request model: {temp_config['model_name']}")

@router.post("/pull_model")
def pull_model_endpoint(payload: PullModelRequest):
    if not payload.hf_id and not payload.safetensors_url:
        raise HTTPException(status_code=400, detail="Provide either 'hf_id' or 'safetensors_url'.")
    
    # 1) Pull Hugging Face model into a folder
    if payload.hf_id:
        model_id = payload.hf_id.strip()
        folder_name = payload.local_name or model_id.replace("/", "__")
        dest_dir = state.models_path / folder_name
        dest_dir.mkdir(parents=True, exist_ok=True)

        try:
            ASCIIColors.cyan(f"Pulling HF model '{model_id}' into {dest_dir}")
            # Reuse config options for HF access
            load_params: Dict[str, Any] = {}
            if state.config.get("hf_cache_path"):
                load_params["cache_dir"] = str(state.config["hf_cache_path"])
            if state.config.get("hf_token"):
                load_params["token"] = state.config["hf_token"]
            # For a pull, we want to actually download:
            load_params["local_files_only"] = False

            # Use DiffusionPipeline (or AutoPipelineForText2Image) to download, then save_pretrained
            pipe = DiffusionPipeline.from_pretrained(model_id, **load_params)
            pipe.save_pretrained(dest_dir)
            del pipe
            gc.collect()
            if torch.cuda.is_available():
                torch.cuda.empty_cache()
            ASCIIColors.green(f"Model '{model_id}' pulled to {dest_dir}")
            return {"status": "ok", "model_name": folder_name}
        except Exception as e:
            trace_exception(e)
            raise HTTPException(status_code=500, detail=f"Failed to pull HF model: {e}")

    # 2) Pull raw .safetensors from URL
    if payload.safetensors_url:
        url = payload.safetensors_url.strip()
        default_name = url.split("/")[-1] or "model.safetensors"
        if not default_name.endswith(".safetensors"):
            default_name += ".safetensors"
        filename = payload.local_name or default_name

        dest_path = state.models_path / filename
        temp_path = dest_path.with_suffix(".temp")

        ASCIIColors.cyan(f"Downloading safetensors from {url} to {dest_path}")
        try:
            with requests.get(url, stream=True) as r:
                r.raise_for_status()
                total_size = int(r.headers.get("content-length", 0))
                with open(temp_path, "wb") as f, tqdm(total=total_size, unit="iB", unit_scale=True, desc=f"Downloading {filename}") as bar:
                    for chunk in r.iter_content(chunk_size=8192):
                        if not chunk:
                            continue
                        f.write(chunk)
                        bar.update(len(chunk))
            shutil.move(temp_path, dest_path)
            ASCIIColors.green(f"Safetensors file downloaded to {dest_path}")
            return {"status": "ok", "model_name": filename}
        except Exception as e:
            if temp_path.exists():
                temp_path.unlink()
            trace_exception(e)
            raise HTTPException(status_code=500, detail=f"Failed to download safetensors: {e}")


@router.get("/list_local_models")
def list_local_models_endpoint():
    local_models = set()
    models_root = Path(args.models_path)
    extra_root = Path(args.extra_models_path) if args.extra_models_path else None

    def scan_root(root: Path):
        if not root or not root.exists():
            return
        
        # 1. Diffusers folders (Recursive)
        for model_index in root.rglob("model_index.json"):
             # For listing just the name, we probably want the folder name or relative path
             # Keeping it simple: folder name.
             local_models.add(model_index.parent.name)

        # 2. Safetensors files (Recursive)
        for safepath in root.rglob("*.safetensors"):
            if (safepath.parent / "model_index.json").exists():
                continue
            local_models.add(safepath.name)
            
    scan_root(models_root)
    scan_root(extra_root)
    
    return sorted(list(local_models))

@app.get("/list_models")
def list_models() -> list[dict]:
    models_root = Path(args.models_path)
    extra_root = Path(args.extra_models_path) if args.extra_models_path else None
    result = []
    seen_paths = set()

    def scan_root(root: Path):
        if not root or not root.exists():
            return
        
        # 1. Diffusers folders (Recursive)
        # We look for model_index.json
        for model_index in root.rglob("model_index.json"):
             folder = model_index.parent
             resolved_path = str(folder.resolve())
             if resolved_path in seen_paths:
                 continue
             seen_paths.add(resolved_path)
             
             result.append({
                "model_name": resolved_path,
                "display_name": folder.name,
                "description": "Local Diffusers pipeline"
             })

        # 2. Safetensors files (Recursive)
        for safepath in root.rglob("*.safetensors"):
            # Skip if part of a diffusers folder
            if (safepath.parent / "model_index.json").exists():
                continue
            
            resolved_path = str(safepath.resolve())
            if resolved_path in seen_paths:
                continue
            seen_paths.add(resolved_path)

            result.append({
                "model_name": resolved_path,
                "display_name": safepath.stem,
                "description": "Local .safetensors checkpoint"
            })

    scan_root(models_root)
    scan_root(extra_root)
    return result




@router.get("/list_available_models")
def list_available_models_endpoint():
    # Use list_models() to get all available models (dicts) then extract names
    models_dicts = list_models()
    discoverable = [m['model_name'] for m in models_dicts]
    return sorted(list(set(discoverable)))

@router.get("/get_settings")
def get_settings_endpoint():
    settings_list = []
    available_models = list_available_models_endpoint()
    schedulers = list(SCHEDULER_MAPPING.keys())
    config_to_display = state.config or state.get_default_config()
    for name, value in config_to_display.items():
        setting = {"name": name, "type": str(type(value).__name__), "value": value}
        if name == "model_name": setting["options"] = available_models
        if name == "scheduler_name": setting["options"] = schedulers
        settings_list.append(setting)
    return settings_list

@router.post("/set_settings")
def set_settings_endpoint(settings: Dict[str, Any]):
    try:
        success = state.update_settings(settings)
        return {"success": success}
    except Exception as e:
        trace_exception(e)
        raise HTTPException(status_code=500, detail=str(e))

@router.get("/status")
def status_endpoint():
    return {"status": "running", "diffusers_available": DIFFUSERS_AVAILABLE, "model_loaded": state.manager.is_loaded if state.manager else False}
    
@router.post("/unload_model")
def unload_model_endpoint():
    if state.manager:
        state.manager._unload_pipeline()
        state.registry.release_manager(state.manager.config)
        state.manager = None
    return {"status": "unloaded"}

@router.get("/ps")
def ps_endpoint():
    managers = state.registry.get_all_managers()
    return [{
            "model_name": m.config.get("model_name"), "is_loaded": m.is_loaded,
            "task": m.current_task, "device": m.config.get("device"), "ref_count": m.ref_count,
            "queue_size": m.queue.qsize(), "last_used": time.ctime(m.last_used_time)
        } for m in managers]

app.include_router(router)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Diffusers TTI Server")
    parser.add_argument("--host", type=str, default="localhost", help="Host to bind to.")
    parser.add_argument("--port", type=int, default=9630, help="Port to bind to.")
    parser.add_argument("--models-path", type=str, required=True, help="Path to the models directory.")
    parser.add_argument("--extra-models-path", type=str, default=None, help="Path to an extra models directory.")
    parser.add_argument(
        "--hf-token",
        type=str,
        default=None,
        help="Optional Hugging Face access token used to download private or gated repos."
    )

    args = parser.parse_args()

    MODELS_PATH = Path(args.models_path)
    EXTRA_MODELS_PATH = Path(args.extra_models_path) if args.extra_models_path else None
    state = ServerState(MODELS_PATH, EXTRA_MODELS_PATH)
    if args.hf_token:
        state.config["hf_token"] = args.hf_token
        ASCIIColors.info("Hugging Face token received via CLI and stored in server config.")
    ASCIIColors.cyan(f"--- Diffusers TTI Server ---")
    ASCIIColors.green(f"Starting server on http://{args.host}:{args.port}")
    ASCIIColors.green(f"Serving models from: {MODELS_PATH.resolve()}")
    if EXTRA_MODELS_PATH:
        ASCIIColors.green(f"Serving extra models from: {EXTRA_MODELS_PATH.resolve()}")
    if not DIFFUSERS_AVAILABLE:
        ASCIIColors.error("Diffusers or its dependencies are not installed correctly in the server's environment!")
    else:
        ASCIIColors.info(f"Detected device: {state.config['device']}")

    uvicorn.run(app, host=args.host, port=args.port, reload=False)



# lollms_client/tti_bindings/lollms/__init__.py
import requests
import base64
import io
from PIL import Image
from lollms_client.lollms_tti_binding import LollmsTTIBinding
from typing import Optional, List, Dict, Any
from ascii_colors import trace_exception, ASCIIColors
import json # Added for potential error parsing

# Defines the binding name for the manager
BindingName = "LollmsWebuiTTIBinding_Impl"
class LollmsWebuiTTIBinding_Impl(LollmsTTIBinding):
    """Concrete implementation of the LollmsTTIBinding for the standard LOLLMS server."""

    def __init__(self,
                 **kwargs):
        """
        Initialize the LOLLMS TTI binding.

        Args:
            host_address (Optional[str]): Host address for the LOLLMS service.
            service_key (Optional[str]): Authentication key (used for client_id verification).
            verify_ssl_certificate (bool): Whether to verify SSL certificates.
        """
        # Prioritize 'model_name' but accept 'model' as an alias from config files.
        if 'model' in kwargs and 'model_name' not in kwargs:
            kwargs['model_name'] = kwargs.pop('model')
        super().__init__(binding_name=BindingName, config=kwargs)

        # Extract parameters from kwargs, providing defaults
        self.host_address = kwargs.get("host_address", "http://localhost:9600")  # Default LOLLMS host
        self.verify_ssl_certificate = kwargs.get("verify_ssl_certificate", True)

        # The 'service_key' here will act as the 'client_id' for TTI requests if provided.
        # This assumes the client library user provides their LOLLMS client_id here.
        self.client_id = kwargs.get("service_key", None) # Use service_key or None

    def _get_client_id(self, **kwargs) -> str:
        """Helper to get client_id, prioritizing kwargs then instance default."""
        client_id = kwargs.get("client_id", self.client_id)
        if not client_id:
             # Allowing anonymous access for generation, but other endpoints might fail
             # raise ValueError("client_id is required for this TTI operation but was not provided.")
             ASCIIColors.warning("client_id not provided for TTI operation. Some features might require it.")
             return "lollms_client_user" # Default anonymous ID
        return client_id

    def generate_image(self,
                       prompt: str,
                       negative_prompt: Optional[str] = "",
                       width: int = 512,
                       height: int = 512,
                       **kwargs) -> bytes:
        """
        Generates image data using the LOLLMS /generate_image endpoint.

        Args:
            prompt (str): The positive text prompt.
            negative_prompt (Optional[str]): The negative prompt.
            width (int): Image width.
            height (int): Image height.
            **kwargs: Additional parameters (passed to server if needed, currently unused by endpoint).

        Returns:
            bytes: The generated image data (JPEG format).

        Raises:
            Exception: If the request fails or image generation fails on the server.
        """
        endpoint = f"{self.host_address}/generate_image"
        request_data = {
            "prompt": prompt,
            "negative_prompt": negative_prompt,
            "width": width,
            "height": height
        }
        # Include extra kwargs if the server API might use them in the future
        request_data.update(kwargs)

        headers = {'Content-Type': 'application/json'}
        # Note: /generate_image endpoint doesn't currently require client_id based on provided code

        try:
            response = requests.post(endpoint, json=request_data, headers=headers, verify=self.verify_ssl_certificate)
            response.raise_for_status()

            response_json = response.json()
            img_base64 = response_json.get("image")

            if not img_base64:
                raise Exception("Server did not return image data.")

            # Decode the base64 string to bytes
            img_bytes = base64.b64decode(img_base64)
            return img_bytes

        except requests.exceptions.RequestException as e:
            trace_exception(e)
            raise Exception(f"HTTP request failed: {e}") from e
        except Exception as e:
            trace_exception(e)
            # Attempt to get error detail from response if possible
            error_detail = "Unknown server error"
            try:
                error_detail = response.json().get("detail", error_detail)
            except: pass
            raise Exception(f"Image generation failed: {e} - Detail: {error_detail}") from e


    def list_services(self, **kwargs) -> List[Dict[str, str]]:
        """
        Lists available TTI services using the LOLLMS /list_tti_services endpoint.

        Args:
            **kwargs: Must include 'client_id' for server verification.

        Returns:
            List[Dict[str, str]]: List of service dictionaries.

        Raises:
            ValueError: If client_id is not provided.
            Exception: If the request fails.
        """
        endpoint = f"{self.host_address}/list_tti_services"
        client_id = self._get_client_id(**kwargs) # Raises ValueError if missing

        request_data = {"client_id": client_id}
        headers = {'Content-Type': 'application/json'}

        try:
            response = requests.post(endpoint, json=request_data, headers=headers, verify=self.verify_ssl_certificate)
            response.raise_for_status()
            return response.json() # Returns the list directly
        except requests.exceptions.RequestException as e:
            trace_exception(e)
            raise Exception(f"HTTP request failed: {e}") from e
        except Exception as e:
            trace_exception(e)
            raise Exception(f"Failed to list TTI services: {e}") from e

    def get_settings(self, **kwargs) -> Optional[Dict[str, Any]]:
        """
        Retrieves TTI settings using the LOLLMS /get_active_tti_settings endpoint.

        Args:
            **kwargs: Must include 'client_id' for server verification.

        Returns:
            Optional[Dict[str, Any]]: Settings dictionary or empty dict if not available/failed.

        Raises:
            ValueError: If client_id is not provided.
            Exception: If the request fails.
        """
        endpoint = f"{self.host_address}/get_active_tti_settings"
        client_id = self._get_client_id(**kwargs) # Raises ValueError if missing

        request_data = {"client_id": client_id}
        headers = {'Content-Type': 'application/json'}

        try:
            response = requests.post(endpoint, json=request_data, headers=headers, verify=self.verify_ssl_certificate)
            response.raise_for_status()
            settings = response.json()
            # The endpoint returns the template directly if successful, or {} if no TTI/settings
            return settings if isinstance(settings, list) else {} # Ensure correct format or empty
        except requests.exceptions.RequestException as e:
            trace_exception(e)
            raise Exception(f"HTTP request failed: {e}") from e
        except Exception as e:
            trace_exception(e)
            raise Exception(f"Failed to get TTI settings: {e}") from e


    def set_settings(self, settings: Dict[str, Any], **kwargs) -> bool:
        """
        Applies TTI settings using the LOLLMS /set_active_tti_settings endpoint.

        Args:
            settings (Dict[str, Any]): The new settings (matching server's expected format).
            **kwargs: Must include 'client_id' for server verification.

        Returns:
            bool: True if successful, False otherwise.

        Raises:
            ValueError: If client_id is not provided.
            Exception: If the request fails.
        """
        endpoint = f"{self.host_address}/set_active_tti_settings"
        client_id = self._get_client_id(**kwargs) # Raises ValueError if missing

        request_data = {
            "client_id": client_id,
            "settings": settings
            }
        headers = {'Content-Type': 'application/json'}

        try:
            response = requests.post(endpoint, json=request_data, headers=headers, verify=self.verify_ssl_certificate)
            response.raise_for_status()
            response_json = response.json()
            return response_json.get("status", False)
        except requests.exceptions.RequestException as e:
            trace_exception(e)
            raise Exception(f"HTTP request failed: {e}") from e
        except Exception as e:
            trace_exception(e)
            raise Exception(f"Failed to set TTI settings: {e}") from e


# lollms_binding.py
from abc import abstractmethod
import importlib
from pathlib import Path
from typing import Optional, Callable, List, Union, Dict, Any
from ascii_colors import trace_exception, ASCIIColors
from lollms_client.lollms_types import MSG_TYPE
from lollms_client.lollms_discussion import LollmsDiscussion
from lollms_client.lollms_utilities import ImageTokenizer
from lollms_client.lollms_base_binding import LollmsBaseBinding
import re
import yaml
import json

def load_known_contexts():
    """
    Loads the known_contexts data from a JSON file.

    Args:
        file_path (str): The path to the JSON file.

    Returns:
        dict: A dictionary containing the known_contexts data, or None if an error occurs.
    """
    try:
        file_path = Path(__file__).parent / "assets" / "models_ctx_sizes.json"
        with open(file_path, "r") as f:
            known_contexts = json.load(f)
        return known_contexts
    except FileNotFoundError:
        print(f"Error: File not found at {file_path}")
        return []
    except json.JSONDecodeError:
        print(f"Error: Could not decode JSON from {file_path}")
        return []
    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        return []

class LollmsLLMBinding(LollmsBaseBinding):
    """Abstract base class for all LOLLMS LLM bindings"""
    
    def __init__(self, 
                 binding_name: Optional[str] ="unknown",
                 **kwargs
        ):
        """
        Initialize the LollmsLLMBinding base class.

        Args:
            binding_name (Optional[str]): The name of the bindingto be used
        """
        super().__init__(binding_name=binding_name, **kwargs)
        self.model_name = None #Must be set by the instance
        self.default_ctx_size = kwargs.get("ctx_size") 
        self.default_n_predict = kwargs.get("n_predict")
        self.default_stream = kwargs.get("stream")
        self.default_temperature = kwargs.get("temperature")
        self.default_top_k = kwargs.get("top_k")
        self.default_top_p = kwargs.get("top_p")
        self.default_repeat_penalty = kwargs.get("repeat_penalty")
        self.default_repeat_last_n = kwargs.get("repeat_last_n")
        self.default_seed = kwargs.get("seed")
        self.default_n_threads = kwargs.get("n_threads")
        self.default_streaming_callback = kwargs.get("streaming_callback")

    
    @abstractmethod
    def generate_text(self,
                    prompt: str,
                    images: Optional[List[str]] = None,
                    system_prompt: str = "",
                    n_predict: Optional[int] = None,
                    stream: Optional[bool] = None,
                    temperature: Optional[float] = None,
                    top_k: Optional[int] = None,
                    top_p: Optional[float] = None,
                    repeat_penalty: Optional[float] = None,
                    repeat_last_n: Optional[int] = None,
                    seed: Optional[int] = None,
                    n_threads: Optional[int] = None,
                    ctx_size: int | None = None,
                    streaming_callback: Optional[Callable[[str, MSG_TYPE], None]] = None,
                    split:Optional[bool]=False, # put to true if the prompt is a discussion
                    user_keyword:Optional[str]="!@>user:",
                    ai_keyword:Optional[str]="!@>assistant:",
                    think: Optional[bool] = False,
                    reasoning_effort: Optional[bool] = "low", # low, medium, high
                    reasoning_summary: Optional[bool] = "auto", # auto
                    **kwargs
                    ) -> Union[str, dict]:
        """
        Generate text using the active LLM binding, using instance defaults if parameters are not provided.
        """
        pass

    def generate_from_messages(self,
                    messages: List[Dict],
                    n_predict: Optional[int] = None,
                    stream: Optional[bool] = None,
                    temperature: Optional[float] = None,
                    top_k: Optional[int] = None,
                    top_p: Optional[float] = None,
                    repeat_penalty: Optional[float] = None,
                    repeat_last_n: Optional[int] = None,
                    seed: Optional[int] = None,
                    n_threads: Optional[int] = None,
                    ctx_size: int | None = None,
                    streaming_callback: Optional[Callable[[str, MSG_TYPE], None]] = None,
                    think: Optional[bool] = False,
                    reasoning_effort: Optional[bool] = "low", # low, medium, high
                    reasoning_summary: Optional[bool] = "auto", # auto
                    **kwargs
                    ) -> Union[str, dict]:
        """
        Generate text using the active LLM binding, using instance defaults if parameters are not provided.
        """
        ASCIIColors.red("This binding does not support generate_from_messages")


    @abstractmethod
    def chat(self,
            discussion: LollmsDiscussion,
            branch_tip_id: Optional[str] = None,
            n_predict: Optional[int] = None,
            stream: Optional[bool] = None,
            temperature: Optional[float] = None,
            top_k: Optional[int] = None,
            top_p: Optional[float] = None,
            repeat_penalty: Optional[float] = None,
            repeat_last_n: Optional[int] = None,
            seed: Optional[int] = None,
            n_threads: Optional[int] = None,
            ctx_size: Optional[int] = None,
            streaming_callback: Optional[Callable[[str, MSG_TYPE], None]] = None,
            think: Optional[bool] = False,
            reasoning_effort: Optional[bool] = "low", # low, medium, high
            reasoning_summary: Optional[bool] = "auto", # auto
            **kwargs
            ) -> Union[str, dict]:
        """
        A method to conduct a chat session with the model using a LollmsDiscussion object.
        """
        pass

    def get_ctx_size(self, model_name: Optional[str|None] = None) -> Optional[int]:
        """
        Retrieves context size for a model from a hardcoded list.
        """
        if model_name is None:
            model_name = self.model_name

        known_contexts = load_known_contexts()

        normalized_model_name = model_name.lower().strip()
        sorted_base_models = sorted(known_contexts.keys(), key=len, reverse=True)

        for base_name in sorted_base_models:
            if base_name in normalized_model_name:
                context_size = known_contexts[base_name]
                ASCIIColors.warning(
                    f"Using hardcoded context size for model '{model_name}' "
                    f"based on base name '{base_name}': {context_size}"
                )
                return context_size

        ASCIIColors.warning(f"Context size not found for model '{model_name}' in the hardcoded list.")
        return None


    @abstractmethod
    def tokenize(self, text: str) -> list:
        """
        Tokenize the input text into a list of tokens.
        """
        pass
    
    @abstractmethod
    def detokenize(self, tokens: list) -> str:
        """
        Convert a list of tokens back to text.
        """
        pass
    
    @abstractmethod
    def count_tokens(self, text: str) -> int:
        """
        Count tokens from a text.
        """        
        pass

    def count_image_tokens(self, image: str) -> int:
        """
        Estimate the number of tokens for an image using ImageTokenizer based on self.model_name.
        """
        try:
            return ImageTokenizer(self.model_name).count_image_tokens(image)
        except Exception as e:
            ASCIIColors.warning(f"Could not estimate image tokens: {e}")
            return -1
    @abstractmethod
    def embed(self, text: str, **kwargs) -> list:
        """
        Get embeddings for the input text using Ollama API
        """
        pass    
    
    @abstractmethod
    def get_model_info(self) -> dict:
        """
        Return information about the current model.
        """
        pass

    def get_zoo(self) -> List[Dict[str, Any]]:
        """
        Returns a list of models available for download.
        each entry is a dict with:
        name, description, size, type, link
        """
        return []

    def download_from_zoo(self, index: int, progress_callback: Callable[[dict], None] = None) -> dict:
        """
        Downloads a model from the zoo using its index.
        """
        return {"status": False, "message": "Not implemented"}

    @abstractmethod
    def load_model(self, model_name: str) -> bool:
        """
        Load a specific model.
        """
        pass


    def split_discussion(self, lollms_prompt_string: str, system_keyword="!@>system:", user_keyword="!@>user:", ai_keyword="!@>assistant:") -> list:
        """
        Splits a LoLLMs prompt into a list of OpenAI-style messages.
        """
        pattern = r"(?={}|{}|{})".format(
            re.escape(system_keyword),
            re.escape(user_keyword),
            re.escape(ai_keyword)
        )
        parts = re.split(pattern, lollms_prompt_string)
        messages = []

        for part in parts:
            part = part.strip()
            if not part:
                continue

            if part.startswith(system_keyword):
                role = "system"
                content = part[len(system_keyword):].strip()
            elif part.startswith(user_keyword):
                role = "user"
                content = part[len(user_keyword):].strip()
            elif part.startswith(ai_keyword):
                role = "assistant"
                content = part[len(ai_keyword):].strip()
            else:
                if not messages:
                    role = "system"
                    content = part
                else:
                    continue

            messages.append({"role": role, "content": content})
            if messages[-1]["content"]=="":
                del messages[-1]
        return messages
    
    def ps(self):
        """
        List models (simulating a process status command).
        Since Lollms/OpenAI API doesn't have a specific 'ps' endpoint for running models with memory stats,
        we list available models and populate structure with available info, leaving hardware stats empty.
        """
        # Since there is no dedicated ps endpoint to see *running* models in the standard OpenAI API,
        # we list available models and try to map relevant info.
        models = self.list_models()
        standardized_models = []
        for m in models:
            standardized_models.append({
                "model_name": m.get("model_name"),
                "size": None,
                "vram_size": None,
                "gpu_usage_percent": None,
                "cpu_usage_percent": None,
                "expires_at": None,
                "parameters_size": None,
                "quantization_level": None,
                "parent_model": None,
                "context_size": m.get("context_length"),
                "owned_by": m.get("owned_by"),
                "created": m.get("created")
            })
        return standardized_models
        
    def get_context_size(self) -> Optional[int]:
        """
        Returns the default context size for the binding.
        """
        return self.default_ctx_size

class LollmsLLMBindingManager:
    """Manages binding discovery and instantiation"""

    def __init__(self, llm_bindings_dir: Union[str, Path] = Path(__file__).parent.parent / "llm_bindings"):
        self.llm_bindings_dir = Path(llm_bindings_dir)
        self.available_bindings = {}

    def _load_binding(self, binding_name: str):
        """Dynamically load a specific binding implementation from the llm bindings directory."""
        binding_dir = self.llm_bindings_dir / binding_name
        if binding_dir.is_dir() and (binding_dir / "__init__.py").exists():
            try:
                module = importlib.import_module(f"lollms_client.llm_bindings.{binding_name}")
                binding_class = getattr(module, module.BindingName)
                self.available_bindings[binding_name] = binding_class
            except Exception as e:
                trace_exception(e)
                print(f"Failed to load binding {binding_name}: {str(e)}")

    def create_binding(self, 
                      binding_name: str,
                      **kwargs) -> Optional[LollmsLLMBinding]:
        """
        Create an instance of a specific binding.
        """
        if binding_name not in self.available_bindings:
            self._load_binding(binding_name)
        
        binding_class = self.available_bindings.get(binding_name)
        if binding_class:
            return binding_class(**kwargs)
        return None
    @staticmethod
    def _get_fallback_description(binding_name: str) -> Dict:
        """
        Generates a default description dictionary for a binding without a description.yaml file.
        """
        return {
            "binding_name": binding_name,
            "title": binding_name.replace("_", " ").title(),
            "author": "Unknown",
            "creation_date": "N/A",
            "last_update_date": "N/A",
            "description": f"A binding for {binding_name}. No description.yaml file was found, so common parameters are shown as a fallback.",
            "input_parameters": [
                {
                    "name": "model_name",
                    "type": "str",
                    "description": "The model name, ID, or filename to be used.",
                    "mandatory": False,
                    "default": ""
                },
                {
                    "name": "host_address",
                    "type": "str",
                    "description": "The host address of the service (for API-based bindings).",
                    "mandatory": False,
                    "default": ""
                },
                {
                    "name": "models_path",
                    "type": "str",
                    "description": "The path to the models directory (for local bindings).",
                    "mandatory": False,
                    "default": ""
                },
                {
                    "name": "service_key",
                    "type": "str",
                    "description": "The API key or service key for authentication (if applicable).",
                    "mandatory": False,
                    "default": ""
                }
            ]
        }

    @staticmethod
    def get_bindings_list(llm_bindings_dir: Union[str, Path]) -> List[Dict]:
        """
        Lists all available LLM bindings by scanning a directory.
        """
        bindings_dir = Path(llm_bindings_dir)
        if not bindings_dir.is_dir():
            return []

        bindings_list = []
        for binding_folder in bindings_dir.iterdir():
            if binding_folder.is_dir() and (binding_folder / "__init__.py").exists():
                binding_name = binding_folder.name
                description_file = binding_folder / "description.yaml"
                
                binding_info = {}
                if description_file.exists():
                    try:
                        with open(description_file, 'r', encoding='utf-8') as f:
                            binding_info = yaml.safe_load(f)
                        binding_info['binding_name'] = binding_name
                    except Exception as e:
                        print(f"Error loading description.yaml for {binding_name}: {e}")
                        binding_info = LollmsLLMBindingManager._get_fallback_description(binding_name)
                else:
                    binding_info = LollmsLLMBindingManager._get_fallback_description(binding_name)
                
                bindings_list.append(binding_info)

        return sorted(bindings_list, key=lambda b: b.get('title', b['binding_name']))
    
    def get_available_bindings(self) -> List[Dict]:
        """
        Retrieves a list of all available LLM bindings with their full descriptions.
        """
        return LollmsLLMBindingManager.get_bindings_list(self.llm_bindings_dir)

def get_available_bindings(llm_bindings_dir: Union[str, Path] = None) -> List[Dict]:
    """
    Lists all available LLM bindings with their detailed descriptions.
    """
    if llm_bindings_dir is None:
        llm_bindings_dir = Path(__file__).parent / "llm_bindings"
    return LollmsLLMBindingManager.get_bindings_list(llm_bindings_dir)

def list_binding_models(llm_binding_name: str, llm_binding_config: Optional[Dict[str, any]]|None = None, llm_bindings_dir: str|Path = Path(__file__).parent / "llm_bindings") -> List[Dict]:
    """
    Lists all available models for a specific binding.
    """
    binding = LollmsLLMBindingManager(llm_bindings_dir).create_binding(
        binding_name=llm_binding_name,
        **{
            k: v
            for k, v in (llm_binding_config or {}).items()
            if k != "binding_name"
        }
    )

    return binding.list_models() if binding else []
